from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from klvm.SExp import SExp


class EvalError(Exception):
    def __init__(self, message: str, sexp: SExp) -> None:
        super().__init__(message)
        self._sexp = sexp



from typing import Callable, List, Optional, Tuple

from .KLVMObject import KLVMObject
from .EvalError import EvalError
from .SExp import CastableType, SExp
from .operators import OperatorDict

from .costs import (
    APPLY_COST,
    QUOTE_COST,
    PATH_LOOKUP_BASE_COST,
    PATH_LOOKUP_COST_PER_LEG,
    PATH_LOOKUP_COST_PER_ZERO_BYTE,
)

OpCallable = Callable[["OpStackType", "ValStackType"], int]
PreOpCallable = Callable[["OpStackType", "ValStackType"], None]
PreEvalFunction = Callable[[SExp, SExp], Optional[Callable[[SExp], object]]]

ValStackType = List[SExp]
OpStackType = List[OpCallable]


def to_pre_eval_op(
    pre_eval_f: PreEvalFunction,
    to_sexp_f: Callable[[CastableType], SExp],
) -> PreOpCallable:
    def my_pre_eval_op(op_stack: OpStackType, value_stack: ValStackType) -> None:
        v = to_sexp_f(value_stack[-1])
        context = pre_eval_f(v.first(), v.rest())
        if callable(context):

            def invoke_context_op(
                op_stack: OpStackType, value_stack: ValStackType
            ) -> int:
                context(to_sexp_f(value_stack[-1]))
                return 0

            op_stack.append(invoke_context_op)

    return my_pre_eval_op


def msb_mask(byte: int) -> int:
    byte |= byte >> 1
    byte |= byte >> 2
    byte |= byte >> 4
    return (byte + 1) >> 1


def run_program(
    program: KLVMObject,
    args: SExp,
    operator_lookup: OperatorDict,
    max_cost: Optional[int] = None,
    pre_eval_f: Optional[PreEvalFunction] = None,
) -> Tuple[int, SExp]:
    _program = SExp.to(program)
    if pre_eval_f is not None:
        pre_eval_op = to_pre_eval_op(pre_eval_f, _program.to)
    else:
        pre_eval_op = None

    def traverse_path(sexp: SExp, env: SExp) -> Tuple[int, SExp]:
        cost = PATH_LOOKUP_BASE_COST
        cost += PATH_LOOKUP_COST_PER_LEG
        if sexp.nullp():
            return cost, sexp.null()

        if sexp.atom is None:
            raise ValueError("Atom must have a non-None atom attribute")
        b = sexp.atom

        end_byte_cursor = 0
        while end_byte_cursor < len(b) and b[end_byte_cursor] == 0:
            end_byte_cursor += 1

        cost += end_byte_cursor * PATH_LOOKUP_COST_PER_ZERO_BYTE
        if end_byte_cursor == len(b):
            return cost, sexp.null()

        # create a bitmask for the most significant *set* bit
        # in the last non-zero byte
        end_bitmask = msb_mask(b[end_byte_cursor])

        byte_cursor = len(b) - 1
        bitmask = 0x01
        while byte_cursor > end_byte_cursor or bitmask < end_bitmask:
            if env.pair is None:
                raise EvalError("path into atom", env)
            if b[byte_cursor] & bitmask:
                env = env.rest()
            else:
                env = env.first()
            cost += PATH_LOOKUP_COST_PER_LEG
            bitmask <<= 1
            if bitmask == 0x100:
                byte_cursor -= 1
                bitmask = 0x01
        return cost, env

    def swap_op(op_stack: OpStackType, value_stack: ValStackType) -> int:
        v2 = value_stack.pop()
        v1 = value_stack.pop()
        value_stack.append(v2)
        value_stack.append(v1)
        return 0

    def cons_op(op_stack: OpStackType, value_stack: ValStackType) -> int:
        v1 = value_stack.pop()
        v2 = value_stack.pop()
        value_stack.append(v1.cons(v2))
        return 0

    def eval_op(op_stack: OpStackType, value_stack: ValStackType) -> int:
        if pre_eval_op:
            pre_eval_op(op_stack, value_stack)

        pair = value_stack.pop()
        sexp = pair.first()
        args = pair.rest()

        # put a bunch of ops on op_stack

        if sexp.pair is None:
            # sexp is an atom
            cost, r = traverse_path(sexp, args)
            value_stack.append(r)
            return cost

        operator = sexp.first()
        if operator.pair:
            from_as_pair = operator.as_pair()
            assert from_as_pair is not None
            new_operator, must_be_nil = from_as_pair
            if new_operator.pair or must_be_nil.atom != b"":
                raise EvalError("in ((X)...) syntax X must be lone atom", sexp)
            new_operand_list = sexp.rest()
            value_stack.append(new_operator)
            value_stack.append(new_operand_list)
            op_stack.append(apply_op)
            return APPLY_COST

        op = operator.as_atom()
        operand_list = sexp.rest()
        if op == operator_lookup.quote_atom:
            value_stack.append(operand_list)
            return QUOTE_COST

        op_stack.append(apply_op)
        value_stack.append(operator)
        while not operand_list.nullp():
            _ = operand_list.first()
            value_stack.append(_.cons(args))
            op_stack.append(cons_op)
            op_stack.append(eval_op)
            op_stack.append(swap_op)
            operand_list = operand_list.rest()
        value_stack.append(operator.null())
        return 1

    def apply_op(op_stack: OpStackType, value_stack: ValStackType) -> int:
        operand_list = value_stack.pop()
        operator = value_stack.pop()
        if operator.pair:
            raise EvalError("internal error", operator)

        op = operator.as_atom()
        assert op is not None
        if op == operator_lookup.apply_atom:
            if operand_list.list_len() != 2:
                raise EvalError("apply requires exactly 2 parameters", operand_list)
            new_program = operand_list.first()
            new_args = operand_list.rest().first()
            value_stack.append(new_program.cons(new_args))
            op_stack.append(eval_op)
            return APPLY_COST

        additional_cost, r = operator_lookup(op, operand_list)
        value_stack.append(r)
        return additional_cost

    op_stack: OpStackType = [eval_op]
    value_stack: ValStackType = [_program.cons(args)]
    cost: int = 0

    while op_stack:
        f = op_stack.pop()
        cost += f(op_stack, value_stack)
        if max_cost and cost > max_cost:
            raise EvalError("cost exceeded", _program.to(max_cost))
    return cost, value_stack[-1]



from typing import Callable, Dict, Generic, Optional, Tuple, TypeVar

import hashlib

from .KLVMObject import KLVMStorage

T = TypeVar("T")


class ObjectCache(Generic[T]):
    """
    `ObjectCache` provides a way to calculate and cache values for each node
    in a klvm object tree. It can be used to calculate the sha256 tree hash
    for an object and save the hash for all the child objects for building
    usage tables, for example.

    It also allows a function that's defined recursively on a klvm tree to
    have a non-recursive implementation (as it keeps a stack of uncached
    objects locally).
    """

    def __init__(self, f: Callable[["ObjectCache[T]", KLVMStorage], Optional[T]]):
        """
        `f`: Callable[ObjectCache, KLVMObject] -> Union[None, T]

        The function `f` is expected to calculate its T value recursively based
        on the T values for the left and right child for a pair. For an atom, the
        function f must calculate the T value directly.

        If a pair is passed and one of the children does not have its T value cached
        in `ObjectCache` yet, return `None` and f will be called with each child in turn.
        Don't recurse in f; that's part of the point of this function.
        """
        self.f = f
        self.lookup: Dict[int, Tuple[T, KLVMStorage]] = dict()

    def get(self, obj: KLVMStorage) -> T:
        obj_id = id(obj)
        if obj_id not in self.lookup:
            obj_list = [obj]
            while obj_list:
                node = obj_list.pop()
                node_id = id(node)
                if node_id not in self.lookup:
                    v = self.f(self, node)
                    if v is None:
                        if node.pair is None:
                            raise ValueError("f returned None for atom", node)
                        obj_list.append(node)
                        obj_list.append(node.pair[0])
                        obj_list.append(node.pair[1])
                    else:
                        self.lookup[node_id] = (v, node)
        return self.lookup[obj_id][0]

    def contains(self, obj: KLVMStorage) -> bool:
        return id(obj) in self.lookup


def treehash(cache: ObjectCache[bytes], obj: KLVMStorage) -> Optional[bytes]:
    """
    This function can be fed to `ObjectCache` to calculate the sha256 tree
    hash for all objects in a tree.
    """
    if obj.pair:
        left, right = obj.pair

        # ensure both `left` and `right` have cached values
        if cache.contains(left) and cache.contains(right):
            left_hash = cache.get(left)
            right_hash = cache.get(right)
            return hashlib.sha256(b"\2" + left_hash + right_hash).digest()
        return None
    assert obj.atom is not None
    return hashlib.sha256(b"\1" + obj.atom).digest()


def serialized_length(cache: ObjectCache[int], obj: KLVMStorage) -> Optional[int]:
    """
    This function can be fed to `ObjectCache` to calculate the serialized
    length for all objects in a tree.
    """
    if obj.pair:
        left, right = obj.pair

        # ensure both `left` and `right` have cached values
        if cache.contains(left) and cache.contains(right):
            left_length = cache.get(left)
            right_length = cache.get(right)
            return 1 + left_length + right_length
        return None
    assert obj.atom is not None
    lb = len(obj.atom)
    if lb == 0 or (lb == 1 and obj.atom[0] < 128):
        return 1
    if lb < 0x40:
        return 1 + lb
    if lb < 0x2000:
        return 2 + lb
    if lb < 0x100000:
        return 3 + lb
    if lb < 0x8000000:
        return 4 + lb
    if lb < 0x400000000:
        return 5 + lb
    raise ValueError("atom of size %d too long" % lb)



from __future__ import annotations

import io
import typing

import typing_extensions

from .as_python import as_python
from .KLVMObject import KLVMObject, KLVMStorage

from .EvalError import EvalError

from .casts import (
    int_from_bytes,
    int_to_bytes,
)
from .serialize import sexp_to_stream


CastableType = typing.Union[
    "SExp",
    KLVMStorage,
    typing.SupportsBytes,
    bytes,
    str,
    int,
    None,
    typing.Sequence["CastableType"],
    typing.Tuple["CastableType", "CastableType"],
]


NULL = b""


def looks_like_klvm_object(o: typing.Any) -> typing_extensions.TypeGuard[KLVMStorage]:
    d = dir(o)
    return "atom" in d and "pair" in d


# this function recognizes some common types and turns them into plain bytes,
def convert_atom_to_bytes(
    v: typing.Union[
        bytes,
        str,
        int,
        None,
        typing.List[typing_extensions.Never],
        typing.SupportsBytes,
    ],
) -> bytes:
    if isinstance(v, bytes):
        return v
    if isinstance(v, str):
        return v.encode()
    if isinstance(v, int):
        return int_to_bytes(v)
    if v is None:
        return b""
    if v == []:
        return b""
    if hasattr(v, "__bytes__"):
        return bytes(v)

    raise ValueError("can't cast %s (%s) to bytes" % (type(v), v))


ValType = typing.Union["SExp", CastableType]
StackType = typing.List[ValType]


# returns a klvm-object like object
@typing.no_type_check
def to_sexp_type(
    v: CastableType,
) -> KLVMStorage:
    stack = [v]
    ops = [(0, None)]  # convert

    while len(ops) > 0:
        op, target = ops.pop()
        # convert value
        if op == 0:
            if looks_like_klvm_object(stack[-1]):
                continue
            v = stack.pop()
            if isinstance(v, tuple):
                if len(v) != 2:
                    raise ValueError("can't cast tuple of size %d" % len(v))
                left, right = v
                target = len(stack)
                stack.append(KLVMObject((left, right)))
                if not looks_like_klvm_object(right):
                    stack.append(right)
                    ops.append((2, target))  # set right
                    ops.append((0, None))  # convert
                if not looks_like_klvm_object(left):
                    stack.append(left)
                    ops.append((1, target))  # set left
                    ops.append((0, None))  # convert
                continue
            if isinstance(v, list):
                target = len(stack)
                stack.append(KLVMObject(NULL))
                for _ in v:
                    stack.append(_)
                    ops.append((3, target))  # prepend list
                    # we only need to convert if it's not already the right
                    # type
                    if not looks_like_klvm_object(_):
                        ops.append((0, None))  # convert
                continue
            stack.append(KLVMObject(convert_atom_to_bytes(v)))
            continue

        if op == 1:  # set left
            stack[target].pair = (KLVMObject(stack.pop()), stack[target].pair[1])
            continue
        if op == 2:  # set right
            stack[target].pair = (stack[target].pair[0], KLVMObject(stack.pop()))
            continue
        if op == 3:  # prepend list
            stack[target] = KLVMObject((stack.pop(), stack[target]))
            continue
    # there's exactly one item left at this point
    if len(stack) != 1:
        raise ValueError("internal error")

    # stack[0] implements the klvm object protocol and can be wrapped by an SExp
    return stack[0]


_T_SExp = typing.TypeVar("_T_SExp", bound="SExp")


class SExp:
    """
    SExp provides higher level API on top of any object implementing the KLVM
    object protocol.
    The tree of values is not a tree of SExp objects, it's a tree of KLVMObject
    like objects. SExp simply wraps them to privide a uniform view of any
    underlying conforming tree structure.

    The KLVM object protocol (concept) exposes two attributes:
    1. "atom" which is either None or bytes
    2. "pair" which is either None or a tuple of exactly two elements. Both
       elements implementing the KLVM object protocol.
    Exactly one of "atom" and "pair" must be None.
    """

    true: typing.ClassVar[SExp]
    false: typing.ClassVar[SExp]
    __null__: typing.ClassVar[SExp]

    # the underlying object implementing the klvm object protocol
    atom: typing.Optional[bytes]

    # this is a tuple of the otherlying KLVMObject-like objects. i.e. not
    # SExp objects with higher level functions, or None
    pair: typing.Optional[typing.Tuple[KLVMStorage, KLVMStorage]]

    def __init__(self, obj: KLVMStorage) -> None:
        self.atom = obj.atom
        self.pair = obj.pair

    # this returns a tuple of two SExp objects, or None
    def as_pair(self) -> typing.Optional[typing.Tuple[SExp, SExp]]:
        pair = self.pair
        if pair is None:
            return pair
        return (self.__class__(pair[0]), self.__class__(pair[1]))

    # TODO: deprecate this. Same as .atom property
    def as_atom(self) -> typing.Optional[bytes]:
        return self.atom

    def listp(self) -> bool:
        return self.pair is not None

    def nullp(self) -> bool:
        v = self.atom
        return v is not None and len(v) == 0

    def as_int(self) -> int:
        if self.atom is None:
            raise TypeError("Unable to convert a pair to an int")
        return int_from_bytes(self.atom)

    def as_bin(self, *, allow_backrefs: bool = False) -> bytes:
        f = io.BytesIO()
        sexp_to_stream(self, f, allow_backrefs=allow_backrefs)
        return f.getvalue()

    # TODO: should be `v: CastableType`
    @classmethod
    def to(cls: typing.Type[_T_SExp], v: typing.Any) -> _T_SExp:
        if isinstance(v, cls):
            return v

        if looks_like_klvm_object(v):
            return cls(v)

        # this will lazily convert elements
        return cls(to_sexp_type(v))

    def cons(self: _T_SExp, right: _T_SExp) -> _T_SExp:
        return self.to((self, right))

    def first(self: _T_SExp) -> _T_SExp:
        pair = self.pair
        if pair:
            return self.__class__(pair[0])
        raise EvalError("first of non-cons", self)

    def rest(self: _T_SExp) -> _T_SExp:
        pair = self.pair
        if pair:
            return self.__class__(pair[1])
        raise EvalError("rest of non-cons", self)

    @classmethod
    def null(class_) -> SExp:
        return class_.__null__

    def as_iter(self: _T_SExp) -> typing.Iterator[_T_SExp]:
        v = self
        while v.pair is not None:
            yield v.first()
            v = v.rest()

    def __eq__(self, other: object) -> bool:
        try:
            other = self.to(typing.cast(CastableType, other))
            to_compare_stack = [(self, other)]
            while to_compare_stack:
                s1, s2 = to_compare_stack.pop()
                p1 = s1.as_pair()
                if p1:
                    p2 = s2.as_pair()
                    if p2:
                        to_compare_stack.append((p1[0], p2[0]))
                        to_compare_stack.append((p1[1], p2[1]))
                    else:
                        return False
                elif s2.as_pair() or s1.as_atom() != s2.as_atom():
                    return False
            return True
        except ValueError:
            return False

    def list_len(self) -> int:
        v = self
        size = 0
        while v.listp():
            size += 1
            v = v.rest()
        return size

    def as_python(self) -> typing.Any:
        return as_python(self)

    def __str__(self) -> str:
        return self.as_bin().hex()

    def __repr__(self) -> str:
        return "%s(%s)" % (self.__class__.__name__, str(self))


SExp.false = SExp.__null__ = SExp(KLVMObject(b""))
SExp.true = SExp(KLVMObject(b"\1"))



from collections import Counter
from typing import Dict, Optional, List, Set, Tuple

import hashlib


LEFT = 0
RIGHT = 1


class ReadCacheLookup:
    """
    When deserializing a klvm object, a stack of deserialized child objects
    is created, which can be used with back-references. A `ReadCacheLookup` keeps
    track of the state of this stack and all child objects under each root
    node in the stack so that we can quickly determine if a relevant
    back-reference is available.

    In other words, if we've already serialized an object with tree hash T,
    and we encounter another object with that tree hash, we don't re-serialize
    it, but rather include a back-reference to it. This data structure lets
    us quickly determine which back-reference has the shortest path.

    Note that there is a counter. This is because the stack contains some
    child objects that are transient, and no longer appear in the stack
    at later times in the parsing. We don't want to waste time looking for
    these objects that no longer exist, so we reference-count them.

    All hashes correspond to sha256 tree hashes.
    """

    def __init__(self) -> None:
        """
        Create a new `ReadCacheLookup` object with just the null terminator
        (ie. an empty list of objects).
        """
        self.root_hash = hashlib.sha256(b"\1").digest()
        self.read_stack: List[Tuple[bytes, bytes]] = []
        self.count: Counter[bytes] = Counter()
        self.parent_paths_for_child: Dict[bytes, List[Tuple[bytes, int]]] = {}

    def push(self, obj_hash: bytes) -> None:
        """
        This function is used to note that an object with the given hash has just
        been pushed to the read stack, and update the lookups as appropriate.
        """
        # we add two new entries: the new root of the tree, and this object (by id)
        # new_root: (obj_hash, old_root)
        new_root_hash = hashlib.sha256(b"\2" + obj_hash + self.root_hash).digest()

        self.read_stack.append((obj_hash, self.root_hash))

        self.count.update([obj_hash, new_root_hash])

        new_parent_to_old_root = (new_root_hash, LEFT)
        self.parent_paths_for_child.setdefault(obj_hash, list()).append(
            new_parent_to_old_root
        )

        new_parent_to_id = (new_root_hash, RIGHT)
        self.parent_paths_for_child.setdefault(self.root_hash, list()).append(
            new_parent_to_id
        )
        self.root_hash = new_root_hash

    def pop(self) -> Tuple[bytes, bytes]:
        """
        This function is used to note that the top object has just been popped
        from the read stack. Return the 2-tuple of the child hashes.
        """
        item = self.read_stack.pop()
        self.count[item[0]] -= 1
        self.count[self.root_hash] -= 1
        self.root_hash = item[1]
        return item

    def pop2_and_cons(self) -> None:
        """
        This function is used to note that a "pop-and-cons" operation has just
        happened. We remove two objects, cons them together, and push the cons,
        updating the internal look-ups as necessary.
        """
        # we remove two items: the right side of each left/right pair
        right = self.pop()
        left = self.pop()

        self.count.update([left[0], right[0]])

        new_root_hash = hashlib.sha256(b"\2" + left[0] + right[0]).digest()

        self.parent_paths_for_child.setdefault(left[0], list()).append(
            (new_root_hash, LEFT)
        )
        self.parent_paths_for_child.setdefault(right[0], list()).append(
            (new_root_hash, RIGHT)
        )
        self.push(new_root_hash)

    def find_paths(self, obj_hash: bytes, serialized_length: int) -> Set[bytes]:
        """
        This function looks for a path from the root to a child node with a given hash
        by using the read cache.
        """
        valid_paths: Set[bytes] = set()
        if serialized_length < 3:
            return valid_paths

        seen_ids: Set[bytes] = set()

        max_bytes_for_path_encoding = serialized_length - 2
        # 1 byte for 0xfe, 1 min byte for savings

        max_path_length = max_bytes_for_path_encoding * 8 - 1
        seen_ids.add(obj_hash)

        partial_paths: List[Tuple[bytes, List[int]]] = [(obj_hash, [])]

        while partial_paths:
            new_seen_ids = set(seen_ids)
            new_partial_paths = []
            for node, path in partial_paths:
                if node == self.root_hash:
                    valid_paths.add(reversed_path_to_bytes(path))
                    continue

                parent_paths = self.parent_paths_for_child.get(node)

                if parent_paths:
                    for parent, direction in parent_paths:
                        if self.count[parent] > 0 and parent not in seen_ids:
                            new_path = list(path)
                            new_path.append(direction)
                            if len(new_path) > max_path_length:
                                return set()
                            new_partial_paths.append((parent, new_path))
                        new_seen_ids.add(parent)
            partial_paths = new_partial_paths
            if valid_paths:
                return valid_paths
            seen_ids = set(new_seen_ids)
        return valid_paths

    def find_path(self, obj_hash: bytes, serialized_length: int) -> Optional[bytes]:
        r = self.find_paths(obj_hash, serialized_length)
        return min(r) if len(r) > 0 else None


def reversed_path_to_bytes(path: List[int]) -> bytes:
    """
    Convert a list of 0/1 (for left/right) values to a path expected by klvm.

    Reverse the list; convert to a binary number; prepend a 1; break into bytes.

    [] => bytes([0b1])
    [0] => bytes([0b10])
    [1] => bytes([0b11])
    [0, 0] => bytes([0b100])
    [0, 1] => bytes([0b101])
    [1, 0] => bytes([0b110])
    [1, 1] => bytes([0b111])
    [0, 0, 1] => bytes([0b1001])
    [1, 1, 1, 1, 0, 0, 0, 0, 1] => bytes([0b11, 0b11100001])
    """

    byte_count = (len(path) + 1 + 7) >> 3
    v = bytearray(byte_count)
    index = byte_count - 1
    mask = 1
    for p in reversed(path):
        if p:
            v[index] |= mask
        if mask == 0x80:
            index -= 1
            mask = 1
        else:
            mask <<= 1
    v[index] |= mask
    return bytes(v)



import unittest

from typing import List, Tuple, Union

from klvm.KLVMObject import KLVMObject
from klvm.SExp import CastableType, SExp
from chik_rs import G1Element
from klvm.EvalError import EvalError


class dummy_class:
    def __init__(self) -> None:
        self.i = 0


def gen_tree(depth: int) -> SExp:
    if depth == 0:
        return SExp.to(1337)
    subtree = gen_tree(depth - 1)
    return SExp.to((subtree, subtree))


fh = bytes.fromhex
H01 = fh("01")
H02 = fh("02")

NestedListOfBytes = Union[bytes, List["NestedListOfBytes"]]
NestedTupleOfBytes = Union[bytes, Tuple["NestedTupleOfBytes", "NestedTupleOfBytes"]]


class AsPythonTest(unittest.TestCase):
    def check_as_python(self, p: CastableType) -> None:
        v = SExp.to(p)
        p1 = v.as_python()
        self.assertEqual(p, p1)

    def test_null(self) -> None:
        self.check_as_python(b"")

    def test_embedded_tuples(self) -> None:
        self.check_as_python((b"10", ((b"200", b"300"), b"400")))

    def test_single_bytes(self) -> None:
        for _ in range(256):
            self.check_as_python(bytes([_]))

    def test_short_lists(self) -> None:
        self.check_as_python(b"")
        for _ in range(256):
            for size in range(1, 5):
                self.check_as_python(bytes([_] * size))

    def test_int(self) -> None:
        v = SExp.to(42)
        self.assertEqual(v.atom, bytes([42]))

    def test_none(self) -> None:
        v = SExp.to(None)
        self.assertEqual(v.atom, b"")

    def test_empty_list(self) -> None:
        v = SExp.to([])
        self.assertEqual(v.atom, b"")

    def test_list_of_one(self) -> None:
        v = SExp.to([1])
        assert v.pair is not None
        self.assertEqual(type(v.pair[0]), KLVMObject)
        self.assertEqual(type(v.pair[1]), KLVMObject)
        from_as_pair = v.as_pair()
        assert from_as_pair is not None
        self.assertEqual(type(from_as_pair[0]), SExp)
        self.assertEqual(type(from_as_pair[1]), SExp)
        self.assertEqual(v.pair[0].atom, b"\x01")
        self.assertEqual(v.pair[1].atom, b"")

    def test_g1element(self) -> None:
        b = fh(
            "b3b8ac537f4fd6bde9b26221d49b54b17a506be147347dae5"
            "d081c0a6572b611d8484e338f3432971a9823976c6a232b"
        )
        v = SExp.to(G1Element.from_bytes(b))
        self.assertEqual(v.atom, b)

    def test_complex(self) -> None:
        self.check_as_python((b"", b"foo"))
        self.check_as_python((b"", b"1"))
        self.check_as_python([b"2", (b"", b"1")])
        self.check_as_python([b"", b"2", (b"", b"1")])
        self.check_as_python(
            [b"", b"1", b"2", [b"30", b"40", b"90"], b"600", (b"", b"18")]
        )

    def test_listp(self) -> None:
        self.assertEqual(SExp.to(42).listp(), False)
        self.assertEqual(SExp.to(b"").listp(), False)
        self.assertEqual(SExp.to(b"1337").listp(), False)

        self.assertEqual(SExp.to((1337, 42)).listp(), True)
        self.assertEqual(SExp.to([1337, 42]).listp(), True)

    def test_nullp(self) -> None:
        self.assertEqual(SExp.to(b"").nullp(), True)
        self.assertEqual(SExp.to(b"1337").nullp(), False)
        self.assertEqual(SExp.to((b"", b"")).nullp(), False)

    def test_constants(self) -> None:
        self.assertEqual(SExp.__null__.nullp(), True)
        self.assertEqual(SExp.null().nullp(), True)
        self.assertEqual(SExp.true, True)
        self.assertEqual(SExp.false, False)

    def test_list_len(self) -> None:
        v = SExp.to(42)
        for i in range(100):
            self.assertEqual(v.list_len(), i)
            v = SExp.to((42, v))
        self.assertEqual(v.list_len(), 100)

    def test_list_len_atom(self) -> None:
        v = SExp.to(42)
        self.assertEqual(v.list_len(), 0)

    def test_as_int(self) -> None:
        self.assertEqual(SExp.to(fh("80")).as_int(), -128)
        self.assertEqual(SExp.to(fh("ff")).as_int(), -1)
        self.assertEqual(SExp.to(fh("0080")).as_int(), 128)
        self.assertEqual(SExp.to(fh("00ff")).as_int(), 255)

    def test_cons(self) -> None:
        # list
        self.assertEqual(
            SExp.to(H01).cons(SExp.to(H02).cons(SExp.null())).as_python(),
            [H01, H02],
        )
        # cons-box of two values
        self.assertEqual(SExp.to(H01).cons(SExp.to(H02).as_python()), (H01, H02))

    def test_string(self) -> None:
        self.assertEqual(SExp.to("foobar").as_atom(), b"foobar")

    def test_deep_recursion(self) -> None:
        d: NestedListOfBytes = b"2"
        for i in range(1000):
            d = [d]
        v = SExp.to(d)
        for i in range(1000):
            from_as_pair = v.as_pair()
            assert from_as_pair is not None
            self.assertEqual(from_as_pair[1].as_atom(), SExp.null())
            v = from_as_pair[0]
            element = d[0]
            assert isinstance(element, (list, bytes))
            d = element

        self.assertEqual(v.as_atom(), b"2")
        self.assertEqual(d, b"2")

    def test_long_linked_list(self) -> None:
        d: NestedTupleOfBytes = b""
        for i in range(1000):
            d = (b"2", d)
        v = SExp.to(d)
        for i in range(1000):
            from_as_pair = v.as_pair()
            assert from_as_pair is not None
            self.assertEqual(from_as_pair[0].as_atom(), d[0])
            from_as_pair = v.as_pair()
            assert from_as_pair is not None
            v = from_as_pair[1]
            assert isinstance(d, tuple)
            d = d[1]

        self.assertEqual(v.as_atom(), SExp.null())
        self.assertEqual(d, b"")

    def test_long_list(self) -> None:
        d = [1337] * 1000
        v = SExp.to(d)
        for i in range(1000 - 1):
            from_as_pair = v.as_pair()
            assert from_as_pair is not None
            self.assertEqual(from_as_pair[0].as_int(), d[i])
            v = from_as_pair[1]

        self.assertEqual(v.as_atom(), SExp.null())

    def test_invalid_type(self) -> None:
        with self.assertRaises(ValueError):
            SExp.to(dummy_class)

    def test_invalid_tuple(self) -> None:
        with self.assertRaises(ValueError):
            SExp.to((dummy_class, dummy_class))

        with self.assertRaises(ValueError):
            SExp.to((dummy_class, dummy_class, dummy_class))

    def test_klvm_object_tuple(self) -> None:
        o1 = KLVMObject(b"foo")
        o2 = KLVMObject(b"bar")
        self.assertEqual(SExp.to((o1, o2)), (o1, o2))

    def test_first(self) -> None:
        val = SExp.to(1)
        self.assertRaises(EvalError, lambda: val.first())
        val = SExp.to((42, val))
        self.assertEqual(val.first(), SExp.to(42))

    def test_rest(self) -> None:
        val = SExp.to(1)
        self.assertRaises(EvalError, lambda: val.rest())
        val = SExp.to((42, val))
        self.assertEqual(val.rest(), SExp.to(1))

    def test_as_iter(self) -> None:
        val = list(SExp.to((1, (2, (3, (4, b""))))).as_iter())
        self.assertEqual(val, [1, 2, 3, 4])

        val = list(SExp.to(b"").as_iter())
        self.assertEqual(val, [])

        val = list(SExp.to((1, b"")).as_iter())
        self.assertEqual(val, [1])

        # we accept lists that are not null-terminated
        val = list(SExp.to(1).as_iter())
        self.assertEqual(val, [])

        val = list(SExp.to((1, (2, (3, (4, 5))))).as_iter())
        self.assertEqual(val, [1, 2, 3, 4])

    def test_eq(self) -> None:
        val = SExp.to(1)

        self.assertTrue(val == 1)
        self.assertFalse(val == 2)

        # mismatching types
        self.assertFalse(val == [1])
        self.assertFalse(val == [1, 2])
        self.assertFalse(val == (1, 2))
        self.assertFalse(val == (dummy_class, dummy_class))

    def test_eq_tree(self) -> None:
        val1 = gen_tree(2)
        val2 = gen_tree(2)
        val3 = gen_tree(3)

        self.assertTrue(val1 == val2)
        self.assertTrue(val2 == val1)
        self.assertFalse(val1 == val3)
        self.assertFalse(val3 == val1)

    def test_str(self) -> None:
        self.assertEqual(str(SExp.to(1)), "01")
        self.assertEqual(str(SExp.to(1337)), "820539")
        self.assertEqual(str(SExp.to(-1)), "81ff")
        self.assertEqual(str(gen_tree(1)), "ff820539820539")
        self.assertEqual(str(gen_tree(2)), "ffff820539820539ff820539820539")

    def test_repr(self) -> None:
        self.assertEqual(repr(SExp.to(1)), "SExp(01)")
        self.assertEqual(repr(SExp.to(1337)), "SExp(820539)")
        self.assertEqual(repr(SExp.to(-1)), "SExp(81ff)")
        self.assertEqual(repr(gen_tree(1)), "SExp(ff820539820539)")
        self.assertEqual(repr(gen_tree(2)), "SExp(ffff820539820539ff820539820539)")



import hashlib
import io
import typing

from chik_rs import G1Element, PrivateKey

from .EvalError import EvalError
from .casts import limbs_for_int
from .SExp import SExp

from .costs import (
    ARITH_BASE_COST,
    ARITH_COST_PER_BYTE,
    ARITH_COST_PER_ARG,
    LOG_BASE_COST,
    LOG_COST_PER_ARG,
    LOG_COST_PER_BYTE,
    DIVMOD_BASE_COST,
    DIVMOD_COST_PER_BYTE,
    DIV_BASE_COST,
    DIV_COST_PER_BYTE,
    MUL_BASE_COST,
    MUL_COST_PER_OP,
    MUL_LINEAR_COST_PER_BYTE,
    MUL_SQUARE_COST_PER_BYTE_DIVIDER,
    SHA256_BASE_COST,
    SHA256_COST_PER_ARG,
    SHA256_COST_PER_BYTE,
    PUBKEY_BASE_COST,
    PUBKEY_COST_PER_BYTE,
    POINT_ADD_BASE_COST,
    POINT_ADD_COST_PER_ARG,
    STRLEN_BASE_COST,
    STRLEN_COST_PER_BYTE,
    CONCAT_BASE_COST,
    CONCAT_COST_PER_ARG,
    CONCAT_COST_PER_BYTE,
    BOOL_BASE_COST,
    BOOL_COST_PER_ARG,
    LOGNOT_BASE_COST,
    LOGNOT_COST_PER_BYTE,
    LSHIFT_BASE_COST,
    LSHIFT_COST_PER_BYTE,
    ASHIFT_BASE_COST,
    ASHIFT_COST_PER_BYTE,
    GRS_BASE_COST,
    GRS_COST_PER_BYTE,
    GR_BASE_COST,
    GR_COST_PER_BYTE,
    MALLOC_COST_PER_BYTE,
)


_T_SExp = typing.TypeVar("_T_SExp", bound=SExp)


def malloc_cost(cost: int, atom: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    if atom.atom is None:
        raise ValueError("Atom must have a non-None atom attribute")
    return cost + len(atom.atom) * MALLOC_COST_PER_BYTE, atom


def op_sha256(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = SHA256_BASE_COST
    arg_len = 0
    h = hashlib.sha256()
    for _ in args.as_iter():
        atom = _.atom
        if atom is None:
            raise EvalError("sha256 on list", _)
        arg_len += len(atom)
        cost += SHA256_COST_PER_ARG
        h.update(atom)
    cost += arg_len * SHA256_COST_PER_BYTE
    return malloc_cost(cost, args.to(h.digest()))


def args_as_ints(op_name: str, args: SExp) -> typing.Iterator[typing.Tuple[int, int]]:
    for arg in args.as_iter():
        if arg.atom is None:
            raise EvalError("%s requires int args" % op_name, arg)
        yield (arg.as_int(), len(arg.atom))


def args_as_int32(op_name: str, args: SExp) -> typing.Iterator[int]:
    for arg in args.as_iter():
        if arg.atom is None:
            raise EvalError("%s requires int32 args" % op_name, arg)
        if len(arg.atom) > 4:
            raise EvalError(
                "%s requires int32 args (with no leading zeros)" % op_name, arg
            )
        yield arg.as_int()


def args_as_int_list(
    op_name: str, args: SExp, count: int
) -> typing.List[typing.Tuple[int, int]]:
    int_list = list(args_as_ints(op_name, args))
    if len(int_list) != count:
        plural = "s" if count != 1 else ""
        raise EvalError(
            "%s takes exactly %d argument%s" % (op_name, count, plural), args
        )
    return int_list


def args_as_bools(op_name: str, args: SExp) -> typing.Iterator[SExp]:
    for arg in args.as_iter():
        v = arg.as_atom()
        if v == b"":
            yield args.false
        else:
            yield args.true


def args_as_bool_list(op_name: str, args: SExp, count: int) -> typing.List[SExp]:
    bool_list = list(args_as_bools(op_name, args))
    if len(bool_list) != count:
        plural = "s" if count != 1 else ""
        raise EvalError(
            "%s takes exactly %d argument%s" % (op_name, count, plural), args
        )
    return bool_list


def op_add(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    total = 0
    cost = ARITH_BASE_COST
    arg_size = 0
    for r, arg_len in args_as_ints("+", args):
        total += r
        arg_size += arg_len
        cost += ARITH_COST_PER_ARG
    cost += arg_size * ARITH_COST_PER_BYTE
    return malloc_cost(cost, args.to(total))


def op_subtract(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = ARITH_BASE_COST
    if args.nullp():
        return malloc_cost(cost, args.to(0))
    sign = 1
    total = 0
    arg_size = 0
    for r, arg_len in args_as_ints("-", args):
        total += sign * r
        sign = -1
        arg_size += arg_len
        cost += ARITH_COST_PER_ARG
    cost += arg_size * ARITH_COST_PER_BYTE
    return malloc_cost(cost, args.to(total))


def op_multiply(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = MUL_BASE_COST
    operands = args_as_ints("*", args)
    try:
        v, vs = next(operands)
    except StopIteration:
        return malloc_cost(cost, args.to(1))

    for r, rs in operands:
        cost += MUL_COST_PER_OP
        cost += (rs + vs) * MUL_LINEAR_COST_PER_BYTE
        cost += (rs * vs) // MUL_SQUARE_COST_PER_BYTE_DIVIDER
        v = v * r
        vs = limbs_for_int(v)
    return malloc_cost(cost, args.to(v))


def op_divmod(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = DIVMOD_BASE_COST
    (i0, l0), (i1, l1) = args_as_int_list("divmod", args, 2)
    if i1 == 0:
        raise EvalError("divmod with 0", args.to(i0))
    cost += (l0 + l1) * DIVMOD_COST_PER_BYTE
    q, r = divmod(i0, i1)
    # since q and r are integers, the atoms will be non-None
    q1_atom: bytes = args.to(q).atom  # type: ignore[assignment]
    r1_atom: bytes = args.to(r).atom  # type: ignore[assignment]
    cost += (len(q1_atom) + len(r1_atom)) * MALLOC_COST_PER_BYTE
    return cost, args.to((q, r))


def op_div(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = DIV_BASE_COST
    (i0, l0), (i1, l1) = args_as_int_list("/", args, 2)
    if i1 == 0:
        raise EvalError("div with 0", args.to(i0))

    if i0 < 0 or i1 < 0:
        raise EvalError("div operator with negative operands is deprecated", args)

    cost += (l0 + l1) * DIV_COST_PER_BYTE
    q, r = divmod(i0, i1)

    return malloc_cost(cost, args.to(q))


def op_gr(args: SExp) -> typing.Tuple[int, SExp]:
    (i0, l0), (i1, l1) = args_as_int_list(">", args, 2)
    cost = GR_BASE_COST
    cost += (l0 + l1) * GR_COST_PER_BYTE
    return cost, args.true if i0 > i1 else args.false


def op_gr_bytes(args: SExp) -> typing.Tuple[int, SExp]:
    arg_list = list(args.as_iter())
    if len(arg_list) != 2:
        raise EvalError(">s takes exactly 2 arguments", args)
    a0, a1 = arg_list
    if a0.pair or a1.pair:
        raise EvalError(">s on list", a0 if a0.pair else a1)
    b0 = a0.as_atom()
    b1 = a1.as_atom()
    cost = GRS_BASE_COST
    if b0 is None or b1 is None:
        raise TypeError("Internal error, both operands must not be None")
    cost += (len(b0) + len(b1)) * GRS_COST_PER_BYTE
    return cost, args.true if b0 > b1 else args.false


def op_pubkey_for_exp(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    ((i0, l0),) = args_as_int_list("pubkey_for_exp", args, 1)
    i0 %= 0x73EDA753299D7D483339D80809A1D80553BDA402FFFE5BFEFFFFFFFF00000001
    exponent = PrivateKey.from_bytes(i0.to_bytes(32, "big"))
    try:
        r = args.to(bytes(exponent.get_g1()))
        cost = PUBKEY_BASE_COST
        cost += l0 * PUBKEY_COST_PER_BYTE
        return malloc_cost(cost, r)
    except Exception as ex:
        raise EvalError("problem in op_pubkey_for_exp: %s" % ex, args)


def op_point_add(items: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = POINT_ADD_BASE_COST
    p = G1Element()

    for _ in items.as_iter():
        if _.pair:
            raise EvalError("point_add on list", _)
        try:
            atom = _.as_atom()
            assert atom is not None
            p += G1Element.from_bytes(atom)
            cost += POINT_ADD_COST_PER_ARG
        except Exception as ex:
            raise EvalError("point_add expects blob, got %s: %s" % (_, ex), items)
    return malloc_cost(cost, items.to(p))


def op_strlen(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    if args.list_len() != 1:
        raise EvalError("strlen takes exactly 1 argument", args)
    a0 = args.first()
    if a0.pair:
        raise EvalError("strlen on list", a0)
    assert a0.atom is not None
    size = len(a0.atom)
    cost = STRLEN_BASE_COST + size * STRLEN_COST_PER_BYTE
    return malloc_cost(cost, args.to(size))


def op_substr(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    arg_count = args.list_len()
    if arg_count not in (2, 3):
        raise EvalError("substr takes exactly 2 or 3 arguments", args)
    a0 = args.first()
    if a0.pair:
        raise EvalError("substr on list", a0)

    s0 = a0.as_atom()
    assert s0 is not None

    if arg_count == 2:
        (i1,) = list(args_as_int32("substr", args.rest()))
        i2 = len(s0)
    else:
        i1, i2 = list(args_as_int32("substr", args.rest()))

    if i2 > len(s0) or i2 < i1 or i2 < 0 or i1 < 0:
        raise EvalError("invalid indices for substr", args)
    s = s0[i1:i2]
    cost = 1
    return cost, args.to(s)


def op_concat(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    cost = CONCAT_BASE_COST
    s = io.BytesIO()
    for arg in args.as_iter():
        if arg.pair:
            raise EvalError("concat on list", arg)
        assert arg.atom is not None
        s.write(arg.atom)
        cost += CONCAT_COST_PER_ARG
    r = s.getvalue()
    cost += len(r) * CONCAT_COST_PER_BYTE
    return malloc_cost(cost, args.to(r))


def op_ash(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    (i0, l0), (i1, l1) = args_as_int_list("ash", args, 2)
    if l1 > 4:
        raise EvalError(
            "ash requires int32 args (with no leading zeros)", args.rest().first()
        )
    if abs(i1) > 65535:
        raise EvalError("shift too large", args.to(i1))
    if i1 >= 0:
        r = i0 << i1
    else:
        r = i0 >> -i1
    cost = ASHIFT_BASE_COST
    cost += (l0 + limbs_for_int(r)) * ASHIFT_COST_PER_BYTE
    return malloc_cost(cost, args.to(r))


def op_lsh(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    (i0, l0), (i1, l1) = args_as_int_list("lsh", args, 2)
    if l1 > 4:
        raise EvalError(
            "lsh requires int32 args (with no leading zeros)", args.rest().first()
        )
    if abs(i1) > 65535:
        raise EvalError("shift too large", args.to(i1))
    # we actually want i0 to be an *unsigned* int
    a0 = args.first().as_atom()
    assert a0 is not None
    i0 = int.from_bytes(a0, "big", signed=False)
    if i1 >= 0:
        r = i0 << i1
    else:
        r = i0 >> -i1
    cost = LSHIFT_BASE_COST
    cost += (l0 + limbs_for_int(r)) * LSHIFT_COST_PER_BYTE
    return malloc_cost(cost, args.to(r))


def binop_reduction(
    op_name: str,
    initial_value: int,
    args: _T_SExp,
    op_f: typing.Callable[[int, int], int],
) -> typing.Tuple[int, _T_SExp]:
    total = initial_value
    arg_size = 0
    cost = LOG_BASE_COST
    for r, arg_len in args_as_ints(op_name, args):
        total = op_f(total, r)
        arg_size += arg_len
        cost += LOG_COST_PER_ARG
    cost += arg_size * LOG_COST_PER_BYTE
    return malloc_cost(cost, args.to(total))


def op_logand(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    def binop(a: int, b: int) -> int:
        a &= b
        return a

    return binop_reduction("logand", -1, args, binop)


def op_logior(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    def binop(a: int, b: int) -> int:
        a |= b
        return a

    return binop_reduction("logior", 0, args, binop)


def op_logxor(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    def binop(a: int, b: int) -> int:
        a ^= b
        return a

    return binop_reduction("logxor", 0, args, binop)


def op_lognot(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    ((i0, l0),) = args_as_int_list("lognot", args, 1)
    cost = LOGNOT_BASE_COST + l0 * LOGNOT_COST_PER_BYTE
    return malloc_cost(cost, args.to(~i0))


def op_not(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    (i0,) = args_as_bool_list("not", args, 1)
    if i0.as_atom() == b"":
        r = args.true
    else:
        r = args.false
    cost = BOOL_BASE_COST
    return cost, args.to(r)


def op_any(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    items = list(args_as_bools("any", args))
    cost = BOOL_BASE_COST + len(items) * BOOL_COST_PER_ARG
    r = args.false
    for v in items:
        if v.as_atom() != b"":
            r = args.true
            break
    return cost, args.to(r)


def op_all(args: _T_SExp) -> typing.Tuple[int, _T_SExp]:
    items = list(args_as_bools("all", args))
    cost = BOOL_BASE_COST + len(items) * BOOL_COST_PER_ARG
    r = args.true
    for v in items:
        if v.as_atom() == b"":
            r = args.false
            break
    return cost, args.to(r)


def op_softfork(args: SExp) -> typing.Tuple[int, SExp]:
    if args.list_len() < 1:
        raise EvalError("softfork takes at least 1 argument", args)
    a = args.first()
    if a.pair:
        raise EvalError("softfork requires int args", a)
    cost = a.as_int()
    if cost < 1:
        raise EvalError("cost must be > 0", args)
    return cost, args.false



from .SExp import SExp
from .operators import (  # noqa
    QUOTE_ATOM,
    KEYWORD_TO_ATOM,
    KEYWORD_FROM_ATOM,
)
from .run_program import run_program  # noqa
from .version import __version__  # noqa

to_sexp_f = SExp.to  # noqa



IF_COST = 33
CONS_COST = 50
FIRST_COST = 30
REST_COST = 30
LISTP_COST = 19

MALLOC_COST_PER_BYTE = 10

ARITH_BASE_COST = 99
ARITH_COST_PER_BYTE = 3
ARITH_COST_PER_ARG = 320

LOG_BASE_COST = 100
LOG_COST_PER_BYTE = 3
LOG_COST_PER_ARG = 264

GRS_BASE_COST = 117
GRS_COST_PER_BYTE = 1

EQ_BASE_COST = 117
EQ_COST_PER_BYTE = 1

GR_BASE_COST = 498
GR_COST_PER_BYTE = 2

DIVMOD_BASE_COST = 1116
DIVMOD_COST_PER_BYTE = 6

DIV_BASE_COST = 988
DIV_COST_PER_BYTE = 4

SHA256_BASE_COST = 87
SHA256_COST_PER_ARG = 134
SHA256_COST_PER_BYTE = 2

POINT_ADD_BASE_COST = 101094
POINT_ADD_COST_PER_ARG = 1343980

PUBKEY_BASE_COST = 1325730
PUBKEY_COST_PER_BYTE = 38

MUL_BASE_COST = 92
MUL_COST_PER_OP = 885
MUL_LINEAR_COST_PER_BYTE = 6
MUL_SQUARE_COST_PER_BYTE_DIVIDER = 128

STRLEN_BASE_COST = 173
STRLEN_COST_PER_BYTE = 1

PATH_LOOKUP_BASE_COST = 40
PATH_LOOKUP_COST_PER_LEG = 4
PATH_LOOKUP_COST_PER_ZERO_BYTE = 4

CONCAT_BASE_COST = 142
CONCAT_COST_PER_ARG = 135
CONCAT_COST_PER_BYTE = 3

BOOL_BASE_COST = 200
BOOL_COST_PER_ARG = 300

ASHIFT_BASE_COST = 596
ASHIFT_COST_PER_BYTE = 3

LSHIFT_BASE_COST = 277
LSHIFT_COST_PER_BYTE = 3

LOGNOT_BASE_COST = 331
LOGNOT_COST_PER_BYTE = 3

APPLY_COST = 90
QUOTE_COST = 20



import gzip
import io
import unittest
from typing import Optional

import pytest

from klvm import to_sexp_f
from klvm.SExp import CastableType, SExp
from klvm.serialize import (
    MAX_SAFE_BYTES,
    _atom_from_stream,
    sexp_from_stream,
    sexp_buffer_from_stream,
    atom_to_byte_iterator,
    sexp_to_stream,
)


TEXT = b"the quick brown fox jumps over the lazy dogs"


class InfiniteStream(io.BytesIO):
    def read(self, n: Optional[int] = -1) -> bytes:
        result = super().read(n)

        if n is not None and n > 0:
            fill_needed = n - len(result)
            result += b" " * fill_needed

        return result


class LargeAtom:
    def __len__(self) -> int:
        return 0x400000001


def has_backrefs(blob: bytes) -> bool:
    """
    Return `True` iff blob has a backref in it.
    """
    f = io.BytesIO(blob)
    obj_count = 1
    while obj_count > 0:
        b = f.read(1)[0]
        if b == 0xFE:
            return True
        if b == 0xFF:
            obj_count += 1
        else:
            _atom_from_stream(f, b)
            obj_count -= 1
    return False


class SerializeTest(unittest.TestCase):
    def check_serde(self, s: CastableType) -> bytes:
        v = to_sexp_f(s)
        b = v.as_bin()
        f = io.BytesIO()
        v1 = sexp_from_stream(io.BytesIO(b), to_sexp_f)
        if v != v1:
            print("%s: %d %r %s" % (v, len(b), b, v1))
            breakpoint()
            b = v.as_bin()
            v1 = sexp_from_stream(io.BytesIO(b), to_sexp_f)
        self.assertEqual(v, v1)
        sexp_to_stream(v1, f)
        length = len(f.getvalue())
        assert f.getbuffer() == b
        # this copies the bytes that represent a single s-expression, just to
        # know where the message ends. It doesn't build a python representaion
        # of it
        buf = sexp_buffer_from_stream(io.BytesIO(b))
        self.assertEqual(buf, b)

        # now turn on backrefs and make sure everything still works

        b2 = v.as_bin(allow_backrefs=True)
        self.assertTrue(len(b2) <= len(b))
        if has_backrefs(b2) or len(b2) < len(b):
            # if we have any backrefs, ensure they actually save space
            self.assertTrue(len(b2) < len(b))
            print(
                "%d bytes before %d after %d saved"
                % (len(b), len(b2), len(b) - len(b2))
            )
            io_b2 = io.BytesIO(b2)
            self.assertRaises(ValueError, lambda: sexp_from_stream(io_b2, to_sexp_f))
            io_b2 = io.BytesIO(b2)
            v2 = sexp_from_stream(io_b2, to_sexp_f, allow_backrefs=True)
            self.assertEqual(v2, s)
            b3 = v2.as_bin()
            self.assertEqual(b, b3)
        with pytest.raises(ValueError, match="SExp exceeds maximum size"):
            f = io.BytesIO()
            sexp_to_stream(v1, f, max_size=length-1)
        f = io.BytesIO()
        sexp_to_stream(v1, f, max_size=length)
        return b2

    def test_zero(self) -> None:
        v = to_sexp_f(b"\x00")
        self.assertEqual(v.as_bin(), b"\x00")

    def test_empty(self) -> None:
        v = to_sexp_f(b"")
        self.assertEqual(v.as_bin(), b"\x80")

    def test_empty_string(self) -> None:
        self.check_serde(b"")

    def test_single_bytes(self) -> None:
        for _ in range(256):
            self.check_serde(bytes([_]))

    def test_short_lists(self) -> None:
        self.check_serde([])
        for _ in range(0, 2048, 8):
            for size in range(1, 5):
                self.check_serde([_] * size)

    def test_cons_box(self) -> None:
        self.check_serde((None, None))
        a: CastableType = (None, 18)
        b: CastableType = [1, 2, 30, 40, 600, a]
        c: CastableType = (None, b)
        self.check_serde(c)
        self.check_serde((100, (TEXT, (30, (50, (90, (TEXT, TEXT + TEXT)))))))

    def test_long_blobs(self) -> None:
        text = TEXT * 300
        for _, t in enumerate(text):
            t1 = text[:_]
            self.check_serde(t1)

    def test_blob_limit(self) -> None:
        with self.assertRaises(ValueError):
            # Specifically substituting another type that is sufficiently similar to
            # the expected bytes for this test.
            for _ in atom_to_byte_iterator(LargeAtom()):  # type: ignore[arg-type]
                pass

    def test_very_long_blobs(self) -> None:
        for size in [0x40, 0x2000, 0x100000, 0x8000000]:
            count = size // len(TEXT)
            text = TEXT * count
            assert len(text) < size
            if len(text) >= MAX_SAFE_BYTES:
                with pytest.raises(ValueError, match="SExp exceeds maximum size"):
                    self.check_serde(text)
            else:
                self.check_serde(text)
            text = TEXT * (count + 1)
            assert len(text) > size
            if len(text) >= MAX_SAFE_BYTES:
                with pytest.raises(ValueError, match="SExp exceeds maximum size"):
                    self.check_serde(text)
            else:
                self.check_serde(text)

    def test_very_deep_tree(self) -> None:
        blob = b"a"
        for depth in [10, 100, 1000, 10000, 100000]:
            s = to_sexp_f(blob)
            for _ in range(depth):
                s = to_sexp_f((s, blob))
            self.check_serde(s)

    def test_deserialize_empty(self) -> None:
        bytes_in = b""
        with self.assertRaises(ValueError):
            sexp_from_stream(io.BytesIO(bytes_in), to_sexp_f)

        with self.assertRaises(ValueError):
            sexp_buffer_from_stream(io.BytesIO(bytes_in))

    def test_deserialize_truncated_size(self) -> None:
        # fe means the total number of bytes in the length-prefix is 7
        # one for each bit set. 5 bytes is too few
        bytes_in = b"\xfe    "
        with self.assertRaises(ValueError):
            sexp_from_stream(io.BytesIO(bytes_in), to_sexp_f)

        with self.assertRaises(ValueError):
            sexp_buffer_from_stream(io.BytesIO(bytes_in))

    def test_deserialize_truncated_blob(self) -> None:
        # this is a complete length prefix. The blob is supposed to be 63 bytes
        # the blob itself is truncated though, it's less than 63 bytes
        bytes_in = b"\xbf   "

        with self.assertRaises(ValueError):
            sexp_from_stream(io.BytesIO(bytes_in), to_sexp_f)

        with self.assertRaises(ValueError):
            sexp_buffer_from_stream(io.BytesIO(bytes_in))

    def test_deserialize_large_blob(self) -> None:
        # this length prefix is 7 bytes long, the last 6 bytes specifies the
        # length of the blob, which is 0xffffffffffff, or (2^48 - 1)
        # we don't support blobs this large, and we should fail immediately when
        # exceeding the max blob size, rather than trying to read this many
        # bytes from the stream
        bytes_in = b"\xfe" + b"\xff" * 6

        with self.assertRaises(ValueError):
            sexp_from_stream(InfiniteStream(bytes_in), to_sexp_f)

        with self.assertRaises(ValueError):
            sexp_buffer_from_stream(InfiniteStream(bytes_in))

    def test_deserialize_generator(self) -> None:
        blob = gzip.GzipFile("tests/generator.bin.gz").read()
        s = sexp_from_stream(io.BytesIO(blob), to_sexp_f)
        b = self.check_serde(s)
        assert len(b) == 19124

    def test_deserialize_bomb(self) -> None:
        def make_bomb(depth: int) -> SExp:
            bomb = to_sexp_f(TEXT)
            for _ in range(depth):
                bomb = to_sexp_f((bomb, bomb))
            return bomb

        bomb_10 = make_bomb(10)
        b10_1 = bomb_10.as_bin(allow_backrefs=False)
        b10_2 = bomb_10.as_bin(allow_backrefs=True)
        self.assertEqual(len(b10_1), 47103)
        self.assertEqual(len(b10_2), 75)

        bomb_20 = make_bomb(20)
        with pytest.raises(ValueError, match="SExp exceeds maximum size"):
            bomb_20.as_bin(allow_backrefs=False)
        b20_2 = bomb_20.as_bin(allow_backrefs=True)
        self.assertEqual(len(b20_2), 105)

        bomb_30 = make_bomb(30)
        with pytest.raises(ValueError, match="SExp exceeds maximum size"):
            bomb_30.as_bin(allow_backrefs=False)
        b30_2 = bomb_30.as_bin(allow_backrefs=True)

        self.assertEqual(len(b30_2), 135)

    def test_specific_tree(self) -> None:
        sexp1 = to_sexp_f((("AAA", "BBB"), ("CCC", "AAA")))
        serialized_sexp1_v1 = sexp1.as_bin(allow_backrefs=False)
        serialized_sexp1_v2 = sexp1.as_bin(allow_backrefs=True)
        self.assertEqual(len(serialized_sexp1_v1), 19)
        self.assertEqual(len(serialized_sexp1_v2), 17)
        deserialized_sexp1_v1 = sexp_from_stream(
            io.BytesIO(serialized_sexp1_v1), to_sexp_f, allow_backrefs=False
        )
        deserialized_sexp1_v2 = sexp_from_stream(
            io.BytesIO(serialized_sexp1_v2), to_sexp_f, allow_backrefs=True
        )
        self.assertTrue(deserialized_sexp1_v1 == deserialized_sexp1_v2)



import unittest

from typing import Dict

from klvm.operators import OperatorProtocol, OperatorDict


class OperatorDictTest(unittest.TestCase):
    def test_operatordict_constructor(self) -> None:
        """Constructing should fail if quote or apply are not specified,
        either by object property or by keyword argument.
        Note that they cannot be specified in the operator dictionary itself.
        """
        # ignoring because apparently it doesn't matter for this test that the types are all wrong
        d: Dict[bytes, OperatorProtocol] = {b"\01": "hello", b"\02": "goodbye"}  # type: ignore [dict-item]
        with self.assertRaises(AssertionError):
            OperatorDict(d)  # type: ignore[call-overload]
        with self.assertRaises(AssertionError):
            OperatorDict(d, apply=b"\01")  # type: ignore[call-overload]
        with self.assertRaises(AssertionError):
            OperatorDict(d, quote=b"\01")  # type: ignore[call-overload]
        o = OperatorDict(d, apply=b"\01", quote=b"\02")
        print(o)
        # Why does the constructed Operator dict contain entries for "apply":1 and "quote":2 ?
        # assert d == o
        self.assertEqual(o.apply_atom, b"\01")
        self.assertEqual(o.quote_atom, b"\02")

        # Test construction from an already existing OperatorDict
        o2 = OperatorDict(o)
        self.assertEqual(o2.apply_atom, b"\01")
        self.assertEqual(o2.quote_atom, b"\02")



from __future__ import annotations

from typing import Dict, Iterator, Optional, Protocol, Tuple, Type, TypeVar, overload

from . import core_ops, more_ops

from .SExp import SExp
from .EvalError import EvalError

from .casts import int_to_bytes
from .op_utils import operators_for_module

from .costs import (
    ARITH_BASE_COST,
    ARITH_COST_PER_BYTE,
    ARITH_COST_PER_ARG,
    MUL_BASE_COST,
    MUL_COST_PER_OP,
    MUL_LINEAR_COST_PER_BYTE,
    MUL_SQUARE_COST_PER_BYTE_DIVIDER,
    CONCAT_BASE_COST,
    CONCAT_COST_PER_ARG,
    CONCAT_COST_PER_BYTE,
)

KEYWORDS = (
    # core opcodes 0x01-x08
    ". q a i c f r l x "
    #
    # opcodes on atoms as strings 0x09-0x0f
    "= >s sha256 substr strlen concat . "
    #
    # opcodes on atoms as ints 0x10-0x17
    "+ - * / divmod > ash lsh "
    #
    # opcodes on atoms as vectors of bools 0x18-0x1c
    "logand logior logxor lognot . "
    #
    # opcodes for bls 1381 0x1d-0x1f
    "point_add pubkey_for_exp . "
    #
    # bool opcodes 0x20-0x23
    "not any all . "
    #
    # misc 0x24
    "softfork "
).split()

KEYWORD_FROM_ATOM = {int_to_bytes(k): v for k, v in enumerate(KEYWORDS)}
KEYWORD_TO_ATOM = {v: k for k, v in KEYWORD_FROM_ATOM.items()}

OP_REWRITE = {
    "+": "add",
    "-": "subtract",
    "*": "multiply",
    "/": "div",
    "i": "if",
    "c": "cons",
    "f": "first",
    "r": "rest",
    "l": "listp",
    "x": "raise",
    "=": "eq",
    ">": "gr",
    ">s": "gr_bytes",
}


def args_len(op_name: str, args: SExp) -> Iterator[int]:
    for arg in args.as_iter():
        if arg.pair:
            raise EvalError("%s requires int args" % op_name, arg)
        # not a pair so must be an atom
        yield len(arg.atom)  # type: ignore[arg-type]


# unknown ops are reserved if they start with 0xffff
# otherwise, unknown ops are no-ops, but they have costs. The cost is computed
# like this:

# byte index (reverse):
# | 4 | 3 | 2 | 1 | 0          |
# +---+---+---+---+------------+
# | multiplier    |XX | XXXXXX |
# +---+---+---+---+---+--------+
#  ^               ^    ^
#  |               |    + 6 bits ignored when computing cost
# cost_multiplier  |
#                  + 2 bits
#                    cost_function

# 1 is always added to the multiplier before using it to multiply the cost, this
# is since cost may not be 0.

# cost_function is 2 bits and defines how cost is computed based on arguments:
# 0: constant, cost is 1 * (multiplier + 1)
# 1: computed like operator add, multiplied by (multiplier + 1)
# 2: computed like operator mul, multiplied by (multiplier + 1)
# 3: computed like operator concat, multiplied by (multiplier + 1)

# this means that unknown ops where cost_function is 1, 2, or 3, may still be
# fatal errors if the arguments passed are not atoms.


def default_unknown_op(op: bytes, args: SExp) -> Tuple[int, SExp]:
    # any opcode starting with ffff is reserved (i.e. fatal error)
    # opcodes are not allowed to be empty
    if len(op) == 0 or op[:2] == b"\xff\xff":
        raise EvalError("reserved operator", args.to(op))

    # all other unknown opcodes are no-ops
    # the cost of the no-ops is determined by the opcode number, except the
    # 6 least significant bits.

    cost_function = (op[-1] & 0b11000000) >> 6
    # the multiplier cannot be 0. it starts at 1

    if len(op) > 5:
        raise EvalError("invalid operator", args.to(op))

    cost_multiplier = int.from_bytes(op[:-1], "big", signed=False) + 1

    # 0 = constant
    # 1 = like op_add/op_sub
    # 2 = like op_multiply
    # 3 = like op_concat
    if cost_function == 0:
        cost = 1
    elif cost_function == 1:
        # like op_add
        cost = ARITH_BASE_COST
        arg_size = 0
        for length in args_len("unknown op", args):
            arg_size += length
            cost += ARITH_COST_PER_ARG
        cost += arg_size * ARITH_COST_PER_BYTE
    elif cost_function == 2:
        # like op_multiply
        cost = MUL_BASE_COST
        operands = args_len("unknown op", args)
        try:
            vs = next(operands)
            for rs in operands:
                cost += MUL_COST_PER_OP
                cost += (rs + vs) * MUL_LINEAR_COST_PER_BYTE
                cost += (rs * vs) // MUL_SQUARE_COST_PER_BYTE_DIVIDER
                # this is an estimate, since we don't want to actually multiply the
                # values
                vs += rs
        except StopIteration:
            pass

    elif cost_function == 3:
        # like concat
        cost = CONCAT_BASE_COST
        length = 0
        for arg in args.as_iter():
            if arg.pair:
                raise EvalError("unknown op on list", arg)
            cost += CONCAT_COST_PER_ARG
            assert arg.atom is not None
            length += len(arg.atom)
        cost += length * CONCAT_COST_PER_BYTE

    cost *= cost_multiplier
    if cost >= 2**32:
        raise EvalError("invalid operator", args.to(op))

    return (cost, SExp.null())


class OperatorProtocol(Protocol):
    def __call__(self, args: SExp) -> Tuple[int, SExp]: ...


class UnknownOperatorProtocol(Protocol):
    def __call__(self, op: bytes, args: SExp) -> Tuple[int, SExp]: ...


_T_OperatorDict = TypeVar("_T_OperatorDict", bound="OperatorDict")


class OperatorDict(Dict[bytes, OperatorProtocol]):
    """
    This is a nice hack that adds `__call__` to a dictionary, so
    operators can be added dynamically.
    """

    unknown_op_handler: UnknownOperatorProtocol
    quote_atom: bytes
    apply_atom: bytes

    @overload
    def __new__(
        cls: Type[_T_OperatorDict],
        d: Dict[bytes, OperatorProtocol],
        quote: bytes,
        apply: bytes,
        unknown_op_handler: UnknownOperatorProtocol = default_unknown_op,
    ) -> _T_OperatorDict: ...

    @overload
    def __new__(
        cls: Type[_T_OperatorDict],
        d: OperatorDict,
        quote: Optional[bytes] = None,
        apply: Optional[bytes] = None,
        unknown_op_handler: UnknownOperatorProtocol = default_unknown_op,
    ) -> _T_OperatorDict: ...

    def __new__(
        cls: Type[_T_OperatorDict],
        d: Dict[bytes, OperatorProtocol],
        quote: Optional[bytes] = None,
        apply: Optional[bytes] = None,
        unknown_op_handler: UnknownOperatorProtocol = default_unknown_op,
    ) -> _T_OperatorDict:
        """
        `quote_atom` and `apply_atom` must be set
        `unknown_op_handler` has a default implementation
        We do not check if quote and apply are distinct
        We do not check if the opcode values for quote and apply exist in the passed-in dict
        """
        self = super().__new__(cls, d)

        if quote is None:
            assert isinstance(d, OperatorDict)
            self.quote_atom = d.quote_atom
        else:
            self.quote_atom = quote

        if apply is None:
            assert isinstance(d, OperatorDict)
            self.apply_atom = d.apply_atom
        else:
            self.apply_atom = apply

        self.unknown_op_handler = unknown_op_handler

        return self

    def __call__(self, op: bytes, arguments: SExp) -> Tuple[int, SExp]:
        f = self.get(op)
        if f is None:
            return self.unknown_op_handler(op, arguments)
        else:
            return f(arguments)


QUOTE_ATOM = KEYWORD_TO_ATOM["q"]
APPLY_ATOM = KEYWORD_TO_ATOM["a"]

OPERATOR_LOOKUP = OperatorDict(
    operators_for_module(KEYWORD_TO_ATOM, core_ops, OP_REWRITE),
    quote=QUOTE_ATOM,
    apply=APPLY_ATOM,
)
OPERATOR_LOOKUP.update(operators_for_module(KEYWORD_TO_ATOM, more_ops, OP_REWRITE))



import unittest

from klvm.run_program import msb_mask


class BitTest(unittest.TestCase):
    def test_msb_mask(self) -> None:
        self.assertEqual(msb_mask(0x0), 0x0)
        self.assertEqual(msb_mask(0x01), 0x01)
        self.assertEqual(msb_mask(0x02), 0x02)
        self.assertEqual(msb_mask(0x04), 0x04)
        self.assertEqual(msb_mask(0x08), 0x08)
        self.assertEqual(msb_mask(0x10), 0x10)
        self.assertEqual(msb_mask(0x20), 0x20)
        self.assertEqual(msb_mask(0x40), 0x40)
        self.assertEqual(msb_mask(0x80), 0x80)

        self.assertEqual(msb_mask(0x44), 0x40)
        self.assertEqual(msb_mask(0x2A), 0x20)
        self.assertEqual(msb_mask(0xFF), 0x80)
        self.assertEqual(msb_mask(0x0F), 0x08)



import unittest

from chik_rs import G1Element, PrivateKey

bls12_381_generator = G1Element.generator()


class BLS12_381_Test(unittest.TestCase):
    def test_stream(self) -> None:
        for _ in range(1, 64):
            p = PrivateKey.from_bytes(_.to_bytes(32, "big")).get_g1()
            blob = bytes(p)
            p1 = G1Element.from_bytes(blob)
            self.assertEqual(len(blob), 48)
            self.assertEqual(p, p1)



import unittest

from klvm import to_sexp_f
from klvm.read_cache_lookup import ReadCacheLookup
from klvm.object_cache import ObjectCache, treehash


class ReadCacheLookupTest(unittest.TestCase):
    def test_various(self) -> None:
        rcl = ReadCacheLookup()
        treehasher = ObjectCache(treehash)

        # rcl = ()
        nil = to_sexp_f(b"")
        nil_hash = treehasher.get(nil)
        self.assertEqual(rcl.root_hash, nil_hash)

        foo = to_sexp_f(b"foo")
        foo_hash = treehasher.get(foo)
        rcl.push(foo_hash)

        # rcl = (foo . 0)

        current_stack = to_sexp_f([foo])
        current_stack_hash = treehasher.get(current_stack)

        self.assertEqual(rcl.root_hash, current_stack_hash)
        self.assertEqual(rcl.find_path(foo_hash, serialized_length=20), bytes([2]))
        self.assertEqual(rcl.find_path(nil_hash, serialized_length=20), bytes([3]))
        self.assertEqual(
            rcl.find_path(current_stack_hash, serialized_length=20), bytes([1])
        )

        bar = to_sexp_f(b"bar")
        bar_hash = treehasher.get(bar)
        rcl.push(bar_hash)

        # rcl = (bar foo)

        current_stack = to_sexp_f([bar, foo])
        current_stack_hash = treehasher.get(current_stack)
        foo_list_hash = treehasher.get(to_sexp_f([b"foo"]))
        self.assertEqual(rcl.root_hash, current_stack_hash)
        self.assertEqual(rcl.find_path(bar_hash, serialized_length=20), bytes([2]))
        self.assertEqual(rcl.find_path(foo_list_hash, serialized_length=20), bytes([3]))
        self.assertEqual(rcl.find_path(foo_hash, serialized_length=20), bytes([5]))
        self.assertEqual(rcl.find_path(nil_hash, serialized_length=20), bytes([7]))
        self.assertEqual(
            rcl.find_path(current_stack_hash, serialized_length=20), bytes([1])
        )
        self.assertEqual(rcl.count[foo_list_hash], 1)

        rcl.pop2_and_cons()
        # rcl = ((foo . bar) . 0)

        current_stack = to_sexp_f([(foo, bar)])
        current_stack_hash = treehasher.get(current_stack)
        self.assertEqual(rcl.root_hash, current_stack_hash)

        # we no longer have `(foo . 0)` in the read stack
        # check that its count is zero
        self.assertEqual(rcl.count[foo_list_hash], 0)

        self.assertEqual(rcl.find_path(bar_hash, serialized_length=20), bytes([6]))
        self.assertEqual(rcl.find_path(foo_list_hash, serialized_length=20), None)
        self.assertEqual(rcl.find_path(foo_hash, serialized_length=20), bytes([4]))
        self.assertEqual(rcl.find_path(nil_hash, serialized_length=20), bytes([3]))
        self.assertEqual(
            rcl.find_path(current_stack_hash, serialized_length=20), bytes([1])
        )

        rcl.push(foo_hash)
        rcl.push(foo_hash)
        rcl.pop2_and_cons()

        # rcl = ((foo . foo) (foo . bar))

        current_stack = to_sexp_f([(foo, foo), (foo, bar)])
        current_stack_hash = treehasher.get(current_stack)
        self.assertEqual(rcl.root_hash, current_stack_hash)
        self.assertEqual(rcl.find_path(bar_hash, serialized_length=20), bytes([13]))
        self.assertEqual(rcl.find_path(foo_list_hash, serialized_length=20), None)
        self.assertEqual(rcl.find_path(foo_hash, serialized_length=20), bytes([4]))
        self.assertEqual(rcl.find_path(nil_hash, serialized_length=20), bytes([7]))

        # find BOTH minimal paths to `foo`
        self.assertEqual(
            rcl.find_paths(foo_hash, serialized_length=20),
            set([bytes([4]), bytes([6])]),
        )

        rcl = ReadCacheLookup()
        rcl.push(foo_hash)
        rcl.push(foo_hash)
        rcl.pop2_and_cons()
        rcl.push(foo_hash)
        rcl.push(foo_hash)
        rcl.pop2_and_cons()
        rcl.pop2_and_cons()
        # rcl = ((foo . foo) . (foo . foo))
        # find ALL minimal paths to `foo`
        self.assertEqual(
            rcl.find_paths(foo_hash, serialized_length=20),
            set([bytes([8]), bytes([10]), bytes([12]), bytes([14])]),
        )



# decoding:
# read a byte
# if it's 0x80, it's nil (which might be same as 0)
# if it's 0xfe, it's a back-reference. Read an atom, and treat it as a path in the cache tree.
# if it's 0xff, it's a cons box. Read two items, build cons
# otherwise, number of leading set bits is length in bytes to read size
# For example, if the bit fields of the first byte read are:
#   10xx xxxx -> 1 byte is allocated for size_byte, and the value of the size is 00xx xxxx
#   110x xxxx -> 2 bytes are allocated for size_byte, and the value of the size 000x xxxx xxxx xxxx
#   1110 xxxx -> 3 bytes allocated. The size is 0000 xxxx xxxx xxxx xxxx xxxx
#   1111 0xxx -> 4 bytes allocated.
#   1111 10xx -> 5 bytes allocated.
# If the first byte read is one of the following:
#   1000 0000 -> 0 bytes : nil
#   0000 0000 -> 1 byte : zero (b'\x00')

import io
import typing

from .read_cache_lookup import ReadCacheLookup
from .object_cache import ObjectCache, treehash, serialized_length

from .KLVMObject import KLVMStorage

MAX_SINGLE_BYTE = 0x7F
BACK_REFERENCE = 0xFE
CONS_BOX_MARKER = 0xFF

MAX_SAFE_BYTES = 2_000_000

T = typing.TypeVar("T")
_T_KLVMStorage = typing.TypeVar("_T_KLVMStorage", bound=KLVMStorage)

CS = typing.TypeVar("CS", bound=KLVMStorage)

ToKLVMStorage = typing.Callable[
    [typing.Union[KLVMStorage, bytes, typing.Tuple[KLVMStorage, KLVMStorage]]],
    _T_KLVMStorage,
]

ValStackType = KLVMStorage

OpCallable = typing.Callable[
    ["OpStackType[T]", ValStackType, typing.BinaryIO, ToKLVMStorage[T]], ValStackType
]

OpStackType = typing.List[OpCallable[T]]


def sexp_to_byte_iterator(
    sexp: KLVMStorage, *, allow_backrefs: bool = False
) -> typing.Iterator[bytes]:
    if allow_backrefs:
        yield from sexp_to_byte_iterator_with_backrefs(sexp)
        return

    todo_stack = [sexp]
    while todo_stack:
        sexp = todo_stack.pop()
        pair = sexp.pair
        if pair:
            yield bytes([CONS_BOX_MARKER])
            todo_stack.append(pair[1])
            todo_stack.append(pair[0])
        else:
            assert sexp.atom is not None
            yield from atom_to_byte_iterator(sexp.atom)


def sexp_to_byte_iterator_with_backrefs(sexp: KLVMStorage) -> typing.Iterator[bytes]:
    # in `read_op_stack`:
    #  "P" = "push"
    #  "C" = "pop two objects, create and push a new cons with them"

    read_op_stack = ["P"]

    write_stack = [sexp]

    read_cache_lookup = ReadCacheLookup()

    thc = ObjectCache(treehash)
    slc = ObjectCache(serialized_length)

    while write_stack:
        node_to_write = write_stack.pop()
        op = read_op_stack.pop()
        assert op == "P"

        node_serialized_length = slc.get(node_to_write)

        node_tree_hash = thc.get(node_to_write)
        path = read_cache_lookup.find_path(node_tree_hash, node_serialized_length)
        if path:
            yield bytes([BACK_REFERENCE])
            yield from atom_to_byte_iterator(path)
            read_cache_lookup.push(node_tree_hash)
        elif node_to_write.pair:
            left, right = node_to_write.pair
            yield bytes([CONS_BOX_MARKER])
            write_stack.append(right)
            write_stack.append(left)
            read_op_stack.append("C")
            read_op_stack.append("P")
            read_op_stack.append("P")
        else:
            atom = node_to_write.atom
            assert atom is not None
            yield from atom_to_byte_iterator(atom)
            read_cache_lookup.push(node_tree_hash)

        while read_op_stack[-1:] == ["C"]:
            read_op_stack.pop()
            read_cache_lookup.pop2_and_cons()


def atom_to_byte_iterator(as_atom: bytes) -> typing.Iterator[bytes]:
    size = len(as_atom)
    if size == 0:
        yield b"\x80"
        return
    if size == 1:
        if as_atom[0] <= MAX_SINGLE_BYTE:
            yield as_atom
            return
    if size < 0x40:
        size_blob = bytes([0x80 | size])
    elif size < 0x2000:
        size_blob = bytes([0xC0 | (size >> 8), (size >> 0) & 0xFF])
    elif size < 0x100000:
        size_blob = bytes([0xE0 | (size >> 16), (size >> 8) & 0xFF, (size >> 0) & 0xFF])
    elif size < 0x8000000:
        size_blob = bytes(
            [
                0xF0 | (size >> 24),
                (size >> 16) & 0xFF,
                (size >> 8) & 0xFF,
                (size >> 0) & 0xFF,
            ]
        )
    elif size < 0x400000000:
        size_blob = bytes(
            [
                0xF8 | (size >> 32),
                (size >> 24) & 0xFF,
                (size >> 16) & 0xFF,
                (size >> 8) & 0xFF,
                (size >> 0) & 0xFF,
            ]
        )
    else:
        raise ValueError(f"sexp too long {as_atom!r}")

    yield size_blob
    yield as_atom


def sexp_to_stream(
    sexp: KLVMStorage, f: typing.BinaryIO, *, allow_backrefs: bool = False, max_size: int = MAX_SAFE_BYTES
) -> None:
    for b in sexp_to_byte_iterator(sexp, allow_backrefs=allow_backrefs):
        max_size -= len(b)
        if max_size < 0:
            raise ValueError("SExp exceeds maximum size")
        f.write(b)


def msb_mask(byte: int) -> int:
    byte |= byte >> 1
    byte |= byte >> 2
    byte |= byte >> 4
    return (byte + 1) >> 1


def traverse_path(
    obj: KLVMStorage, path: bytes, to_sexp: ToKLVMStorage[CS]
) -> KLVMStorage:
    path_as_int = int.from_bytes(path, "big")
    if path_as_int == 0:
        return to_sexp(b"")

    while path_as_int > 1:
        if obj.pair is None:
            raise ValueError("path into atom", obj)
        obj = obj.pair[path_as_int & 1]
        path_as_int >>= 1

    return obj


def _op_cons(
    op_stack: OpStackType[CS],
    val_stack: ValStackType,
    f: typing.BinaryIO,
    to_sexp: ToKLVMStorage[CS],
) -> ValStackType:
    assert val_stack.pair is not None
    right, val_stack = val_stack.pair
    assert val_stack.pair is not None
    left, val_stack = val_stack.pair
    new_cons = to_sexp((left, right))
    return to_sexp((new_cons, val_stack))


def _op_read_sexp(
    op_stack: OpStackType[CS],
    val_stack: ValStackType,
    f: typing.BinaryIO,
    to_sexp: ToKLVMStorage[CS],
) -> ValStackType:
    blob = f.read(1)
    if len(blob) == 0:
        raise ValueError("bad encoding")
    b = blob[0]
    if b == CONS_BOX_MARKER:
        op_stack.append(_op_cons)
        op_stack.append(_op_read_sexp)
        op_stack.append(_op_read_sexp)
        return val_stack
    atom_as_sexp = to_sexp(_atom_from_stream(f, b))
    return to_sexp((atom_as_sexp, val_stack))


def _op_read_sexp_allow_backrefs(
    op_stack: OpStackType[CS],
    val_stack: ValStackType,
    f: typing.BinaryIO,
    to_sexp: ToKLVMStorage[CS],
) -> KLVMStorage:
    blob = f.read(1)
    if len(blob) == 0:
        raise ValueError("bad encoding")
    b = blob[0]
    if b == CONS_BOX_MARKER:
        op_stack.append(_op_cons)
        op_stack.append(_op_read_sexp_allow_backrefs)
        op_stack.append(_op_read_sexp_allow_backrefs)
        return val_stack
    if b == BACK_REFERENCE:
        blob = f.read(1)
        if len(blob) == 0:
            raise ValueError("bad encoding")
        path = _atom_from_stream(f, blob[0])
        backref = traverse_path(val_stack, path, to_sexp)
        return to_sexp((backref, val_stack))
    atom_as_sexp = to_sexp(_atom_from_stream(f, b))
    return to_sexp((atom_as_sexp, val_stack))


def sexp_from_stream(
    f: typing.BinaryIO, to_sexp: ToKLVMStorage[CS], *, allow_backrefs: bool = False
) -> CS:
    op_stack: OpStackType[CS] = [
        _op_read_sexp_allow_backrefs if allow_backrefs else _op_read_sexp
    ]
    val_stack: ValStackType = to_sexp(b"")

    while op_stack:
        func = op_stack.pop()
        val_stack = func(op_stack, val_stack, f, to_sexp)
    assert val_stack.pair is not None
    return to_sexp(val_stack.pair[0])


def _op_consume_sexp(f: typing.BinaryIO) -> typing.Tuple[bytes, int]:
    blob = f.read(1)
    if len(blob) == 0:
        raise ValueError("bad encoding")
    b = blob[0]
    if b == CONS_BOX_MARKER:
        return (blob, 2)
    return (_consume_atom(f, b), 0)


def _consume_atom(f: typing.BinaryIO, b: int) -> bytes:
    if b == 0x80:
        return bytes([b])
    if b <= MAX_SINGLE_BYTE:
        return bytes([b])
    bit_count = 0
    bit_mask = 0x80
    ll = b
    while ll & bit_mask:
        bit_count += 1
        ll &= 0xFF ^ bit_mask
        bit_mask >>= 1
    size_blob = bytes([ll])
    if bit_count > 1:
        llb = f.read(bit_count - 1)
        if len(llb) != bit_count - 1:
            raise ValueError("bad encoding")
        size_blob += llb
    size = int.from_bytes(size_blob, "big")
    if size >= 0x400000000:
        raise ValueError("blob too large")
    blob = f.read(size)
    if len(blob) != size:
        raise ValueError("bad encoding")
    return bytes([b]) + size_blob[1:] + blob


# instead of parsing the input stream, this function pulls out all the bytes
# that represent on S-expression tree, and returns them. This is more efficient
# than parsing and returning a python S-expression tree.
def sexp_buffer_from_stream(f: typing.BinaryIO) -> bytes:
    ret = io.BytesIO()

    depth = 1
    while depth > 0:
        depth -= 1
        buf, d = _op_consume_sexp(f)
        depth += d
        ret.write(buf)
    return ret.getvalue()


def _atom_from_stream(f: typing.BinaryIO, b: int) -> bytes:
    if b == 0x80:
        return b""
    if b <= MAX_SINGLE_BYTE:
        return bytes([b])
    bit_count = 0
    bit_mask = 0x80
    while b & bit_mask:
        bit_count += 1
        b &= 0xFF ^ bit_mask
        bit_mask >>= 1
    size_blob = bytes([b])
    if bit_count > 1:
        bb = f.read(bit_count - 1)
        if len(bb) != bit_count - 1:
            raise ValueError("bad encoding")
        size_blob += bb
    size = int.from_bytes(size_blob, "big")
    if size >= 0x400000000:
        raise ValueError("blob too large")
    blob = f.read(size)
    if len(blob) != size:
        raise ValueError("bad encoding")
    return blob



from __future__ import annotations

import typing


class KLVMStorage(typing.Protocol):
    # It's not clear if it is possible to express the exclusivity without maybe
    # restructuring all the classes, such as having a separate instance for each
    # of the atom and pair cases and hinting a union of a protocol of each type.
    atom: typing.Optional[bytes]
    pair: typing.Optional[PairType]


PairType = typing.Tuple[KLVMStorage, KLVMStorage]


_T_KLVMObject = typing.TypeVar("_T_KLVMObject", bound="KLVMObject")


class KLVMObject:
    """
    This class implements the KLVM Object protocol in the simplest possible way,
    by just having an "atom" and a "pair" field
    """

    atom: typing.Optional[bytes]

    # this is always a 2-tuple of an object implementing the KLVM object
    # protocol.
    pair: typing.Optional[PairType]
    __slots__ = ["atom", "pair"]

    @staticmethod
    def __new__(
        class_: typing.Type[_T_KLVMObject],
        v: typing.Union[_T_KLVMObject, bytes, PairType],
    ) -> _T_KLVMObject:
        if isinstance(v, class_):
            return v
        # mypy does not realize that the isinstance check is type narrowing like this
        narrowed_v: typing.Union[bytes, PairType] = v  # type: ignore[assignment]

        self = super(KLVMObject, class_).__new__(class_)
        if isinstance(narrowed_v, tuple):
            if len(narrowed_v) != 2:
                raise ValueError(
                    "tuples must be of size 2, cannot create KLVMObject from: %s"
                    % str(narrowed_v)
                )
            self.pair = narrowed_v
            self.atom = None
        else:
            self.atom = narrowed_v
            self.pair = None
        return self



from __future__ import annotations

import types
from typing import Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from klvm.operators import OperatorProtocol


def operators_for_dict(
    keyword_to_atom: Dict[str, bytes],
    op_dict: Dict[str, OperatorProtocol],
    op_name_lookup: Optional[Dict[str, str]] = None,
) -> Dict[bytes, OperatorProtocol]:
    if op_name_lookup is None:
        op_name_lookup = {}

    d = {}
    for op in keyword_to_atom.keys():
        op_name = "op_%s" % op_name_lookup.get(op, op)
        op_f = op_dict.get(op_name)
        if op_f:
            d[keyword_to_atom[op]] = op_f
    return d


def operators_for_module(
    keyword_to_atom: Dict[str, bytes],
    mod: types.ModuleType,
    op_name_lookup: Optional[Dict[str, str]] = None,
) -> Dict[bytes, OperatorProtocol]:
    if op_name_lookup is None:
        op_name_lookup = {}
    return operators_for_dict(keyword_to_atom, mod.__dict__, op_name_lookup)



import unittest
from typing import Tuple

from klvm.operators import (
    OPERATOR_LOOKUP,
    KEYWORD_TO_ATOM,
    default_unknown_op,
    OperatorDict,
)
from klvm.EvalError import EvalError
from klvm.SExp import SExp
from klvm.costs import CONCAT_BASE_COST


class OperatorsTest(unittest.TestCase):
    def setUp(self) -> None:
        self.handler_called = False

    def unknown_handler(self, name: bytes, args: SExp) -> Tuple[int, SExp]:
        self.handler_called = True
        self.assertEqual(name, b"\xff\xff1337")
        self.assertEqual(args, SExp.to(1337))
        return 42, SExp.to(b"foobar")

    def test_unknown_op(self) -> None:
        self.assertRaises(
            EvalError, lambda: OPERATOR_LOOKUP(b"\xff\xff1337", SExp.to(1337))
        )
        od = OperatorDict(
            OPERATOR_LOOKUP,
            unknown_op_handler=lambda name, args: self.unknown_handler(name, args),
        )
        cost, ret = od(b"\xff\xff1337", SExp.to(1337))
        self.assertTrue(self.handler_called)
        self.assertEqual(cost, 42)
        self.assertEqual(ret, SExp.to(b"foobar"))

    def test_plus(self) -> None:
        print(OPERATOR_LOOKUP)
        self.assertEqual(
            OPERATOR_LOOKUP(KEYWORD_TO_ATOM["+"], SExp.to([3, 4, 5]))[1], SExp.to(12)
        )

    def test_unknown_op_reserved(self) -> None:
        # any op that starts with ffff is reserved, and results in a hard
        # failure
        with self.assertRaises(EvalError):
            default_unknown_op(b"\xff\xff", SExp.null())

        for suffix in [b"\xff", b"0", b"\x00", b"\xcc\xcc\xfe\xed\xfa\xce"]:
            with self.assertRaises(EvalError):
                default_unknown_op(b"\xff\xff" + suffix, SExp.null())

        with self.assertRaises(EvalError):
            # an empty atom is not a valid opcode
            default_unknown_op(b"", SExp.null())

        # a single ff is not sufficient to be treated as a reserved opcode
        self.assertEqual(
            default_unknown_op(b"\xff", SExp.null()), (CONCAT_BASE_COST, SExp.null())
        )

        # leading zeroes count, and this does not count as a ffff-prefix
        # the cost is 0xffff00 = 16776960
        self.assertEqual(
            default_unknown_op(b"\x00\xff\xff\x00\x00", SExp.null()),
            (16776961, SExp.null()),
        )

    def test_unknown_ops_last_bits(self) -> None:
        # The last byte is ignored for no-op unknown ops
        for suffix in [b"\x3f", b"\x0f", b"\x00", b"\x2c"]:
            # the cost is unchanged by the last byte
            self.assertEqual(
                default_unknown_op(b"\x3c" + suffix, SExp.null()), (61, SExp.null())
            )



import unittest

from klvm.object_cache import ObjectCache, treehash, serialized_length

from klvm_tools.binutils import assemble


class ObjectCacheTest(unittest.TestCase):
    def check(self, obj_text: str, expected_hash: str, expected_length: int) -> None:
        obj = assemble(obj_text)
        th = ObjectCache(treehash)
        self.assertEqual(th.get(obj).hex(), expected_hash)
        sl = ObjectCache(serialized_length)
        self.assertEqual(sl.get(obj), expected_length)

    def test_various(self) -> None:
        self.check(
            "0x00",
            "47dc540c94ceb704a23875c11273e16bb0b8a87aed84de911f2133568115f254",
            1,
        )

        self.check(
            "0", "4bf5122f344554c53bde2ebb8cd2b7e3d1600ad631c385a5d7cce23c7785459a", 1
        )

        self.check(
            "foo", "0080b50a51ecd0ccfaaa4d49dba866fe58724f18445d30202bafb03e21eef6cb", 4
        )

        self.check(
            "(foo . bar)",
            "c518e45ae6a7b4146017b7a1d81639051b132f1f5572ce3088a3898a9ed1280b",
            9,
        )

        self.check(
            "(this is a longer test of a deeper tree)",
            "0a072d7d860d77d8e290ced0fdb29a271198ca3db54d701c45d831e3aae6422c",
            47,
        )



def int_from_bytes(blob: bytes) -> int:
    size = len(blob)
    if size == 0:
        return 0
    return int.from_bytes(blob, "big", signed=True)


def int_to_bytes(v: int) -> bytes:
    byte_count = (v.bit_length() + 8) >> 3
    if v == 0:
        return b""
    r = v.to_bytes(byte_count, "big", signed=True)
    # make sure the string returned is minimal
    # ie. no leading 00 or ff bytes that are unnecessary
    while len(r) > 1 and r[0] == (0xFF if r[1] & 0x80 else 0):
        r = r[1:]
    return r


def limbs_for_int(v: int) -> int:
    """
    Return the number of bytes required to represent this integer.
    """
    return (v.bit_length() + 7) >> 3



import unittest
from dataclasses import dataclass

from typing import ClassVar, Optional, TYPE_CHECKING, Tuple, cast
from klvm.SExp import SExp, looks_like_klvm_object, convert_atom_to_bytes
from klvm.KLVMObject import KLVMObject, KLVMStorage


def validate_sexp(sexp: SExp) -> None:
    validate_stack = [sexp]
    while validate_stack:
        v = validate_stack.pop()
        assert isinstance(v, SExp)
        if v.pair:
            assert isinstance(v.pair, tuple)
            v1, v2 = v.pair
            assert looks_like_klvm_object(v1)
            assert looks_like_klvm_object(v2)
            from_as_pair = v.as_pair()
            assert from_as_pair is not None
            s1, s2 = from_as_pair
            validate_stack.append(s1)
            validate_stack.append(s2)
        else:
            assert isinstance(v.atom, bytes)


def print_leaves(tree: SExp) -> str:
    a = tree.as_atom()
    if a is not None:
        if len(a) == 0:
            return "() "
        return "%d " % a[0]

    ret = ""
    from_as_pair = tree.as_pair()
    assert from_as_pair is not None
    for i in from_as_pair:
        ret += print_leaves(i)

    return ret


def print_tree(tree: SExp) -> str:
    a = tree.as_atom()
    if a is not None:
        if len(a) == 0:
            return "() "
        return "%d " % a[0]

    ret = "("
    from_as_pair = tree.as_pair()
    assert from_as_pair is not None
    for i in from_as_pair:
        ret += print_tree(i)
    ret += ")"
    return ret


@dataclass(frozen=True)
class PairAndAtom:
    pair: None = None
    atom: None = None


@dataclass(frozen=True)
class Pair:
    pair: None = None


@dataclass(frozen=True)
class Atom:
    atom: None = None


class DummyByteConvertible:
    def __bytes__(self) -> bytes:
        return b"foobar"


class ToSExpTest(unittest.TestCase):
    def test_cast_1(self) -> None:
        # this was a problem in `klvm_tools` and is included
        # to prevent regressions
        sexp = SExp.to(b"foo")
        t1 = sexp.to([1, sexp])
        validate_sexp(t1)

    def test_wrap_sexp(self) -> None:
        # it's a bit of a layer violation that KLVMObject unwraps SExp, but we
        # rely on that in a fair number of places for now. We should probably
        # work towards phasing that out

        # making sure this works despite hinting against it at KLVMObject
        o = KLVMObject(SExp.to(1))  # type: ignore[arg-type]
        assert o.atom == bytes([1])

    def test_arbitrary_underlying_tree(self) -> None:
        # SExp provides a view on top of a tree of arbitrary types, as long as
        # those types implement the KLVMObject protocol. This is an example of
        # a tree that's generated
        class GeneratedTree:
            if TYPE_CHECKING:
                _type_check: ClassVar[KLVMStorage] = cast("GeneratedTree", None)

            depth: int = 4
            val: int = 0

            def __init__(self, depth: int, val: int) -> None:
                assert depth >= 0
                self.depth = depth
                self.val = val

            @property
            def atom(self) -> Optional[bytes]:
                if self.depth > 0:
                    return None
                return bytes([self.val])

            @atom.setter
            def atom(self, val: Optional[bytes]) -> None:
                raise RuntimeError("setting not supported in this test class")

            @property
            def pair(self) -> Optional[Tuple[KLVMStorage, KLVMStorage]]:
                if self.depth == 0:
                    return None
                new_depth: int = self.depth - 1
                return (
                    GeneratedTree(new_depth, self.val),
                    GeneratedTree(new_depth, self.val + 2**new_depth),
                )

            @pair.setter
            def pair(self, val: Optional[Tuple[KLVMStorage, KLVMStorage]]) -> None:
                raise RuntimeError("setting not supported in this test class")

        tree = SExp.to(GeneratedTree(5, 0))
        assert (
            print_leaves(tree)
            == "0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 "
            + "16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 "
        )

        tree = SExp.to(GeneratedTree(3, 0))
        assert print_leaves(tree) == "0 1 2 3 4 5 6 7 "

        tree = SExp.to(GeneratedTree(3, 10))
        assert print_leaves(tree) == "10 11 12 13 14 15 16 17 "

    def test_looks_like_klvm_object(self) -> None:
        # this function can't look at the values, that would cause a cascade of
        # eager evaluation/conversion
        pair_and_atom = PairAndAtom()
        print(dir(pair_and_atom))
        assert looks_like_klvm_object(pair_and_atom)

        pair = Pair()
        assert not looks_like_klvm_object(pair)

        atom = Atom()
        assert not looks_like_klvm_object(atom)

    def test_list_conversions(self) -> None:
        a = SExp.to([1, 2, 3])
        assert print_tree(a) == "(1 (2 (3 () )))"

    def test_string_conversions(self) -> None:
        a = SExp.to("foobar")
        assert a.as_atom() == "foobar".encode()

    def test_int_conversions(self) -> None:
        a = SExp.to(1337)
        assert a.as_atom() == bytes([0x5, 0x39])

    def test_none_conversions(self) -> None:
        a = SExp.to(None)
        assert a.as_atom() == b""

    def test_empty_list_conversions(self) -> None:
        a = SExp.to([])
        assert a.as_atom() == b""

    def test_eager_conversion(self) -> None:
        with self.assertRaises(ValueError):
            SExp.to(("foobar", (1, {})))

    def test_convert_atom(self) -> None:
        assert convert_atom_to_bytes(0x133742) == bytes([0x13, 0x37, 0x42])
        assert convert_atom_to_bytes(0x833742) == bytes([0x00, 0x83, 0x37, 0x42])
        assert convert_atom_to_bytes(0) == b""

        assert convert_atom_to_bytes("foobar") == "foobar".encode()
        assert convert_atom_to_bytes("") == b""

        assert convert_atom_to_bytes(b"foobar") == b"foobar"
        assert convert_atom_to_bytes(None) == b""
        assert convert_atom_to_bytes([]) == b""

        assert convert_atom_to_bytes(DummyByteConvertible()) == b"foobar"

        with self.assertRaises(ValueError):
            convert_atom_to_bytes([1, 2, 3])  # type: ignore[arg-type]

        with self.assertRaises(ValueError):
            convert_atom_to_bytes((1, 2))  # type: ignore[arg-type]

        with self.assertRaises(ValueError):
            convert_atom_to_bytes({})  # type: ignore[arg-type]



#!/usr/bin/env python

from setuptools import setup

with open("README.md", "rt") as fh:
    long_description = fh.read()

dependencies = [
    "chik_rs>=0.2.13",
    "importlib_metadata~=7.1",
    "typing-extensions~=4.0",
]

dev_dependencies = [
    "klvm_tools>=0.4.4",
    "mypy",
    "pytest",
    "setuptools",
    "types-setuptools",
]

setup(
    name="klvm",
    packages=[
        "klvm",
    ],
    author="Chik Network, Inc.",
    author_email="hello@chiknetwork.com",
    url="https://github.com/Chik-Network/klvm",
    license="https://opensource.org/licenses/Apache-2.0",
    description="[Contract Language | Chiklisp] Virtual Machine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8.1, <4",
    install_requires=dependencies,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Security :: Cryptography",
    ],
    extras_require=dict(
        dev=dev_dependencies,
    ),
    project_urls={
        "Bug Reports": "https://github.com/Chik-Network/klvm",
        "Source": "https://github.com/Chik-Network/klvm",
    },
    package_data={
        "": ["py.typed"],
    },
)



from typing import Tuple, TypeVar

from typing_extensions import Never

from .EvalError import EvalError
from .SExp import SExp

from .costs import (
    IF_COST,
    CONS_COST,
    FIRST_COST,
    REST_COST,
    LISTP_COST,
    EQ_BASE_COST,
    EQ_COST_PER_BYTE,
)


_T_SExp = TypeVar("_T_SExp", bound=SExp)


def op_if(args: _T_SExp) -> Tuple[int, _T_SExp]:
    if args.list_len() != 3:
        raise EvalError("i takes exactly 3 arguments", args)
    r = args.rest()
    if args.first().nullp():
        return IF_COST, r.rest().first()
    return IF_COST, r.first()


def op_cons(args: _T_SExp) -> Tuple[int, _T_SExp]:
    if args.list_len() != 2:
        raise EvalError("c takes exactly 2 arguments", args)
    return CONS_COST, args.first().cons(args.rest().first())


def op_first(args: _T_SExp) -> Tuple[int, _T_SExp]:
    if args.list_len() != 1:
        raise EvalError("f takes exactly 1 argument", args)
    return FIRST_COST, args.first().first()


def op_rest(args: _T_SExp) -> Tuple[int, _T_SExp]:
    if args.list_len() != 1:
        raise EvalError("r takes exactly 1 argument", args)
    return REST_COST, args.first().rest()


def op_listp(args: SExp) -> Tuple[int, SExp]:
    if args.list_len() != 1:
        raise EvalError("l takes exactly 1 argument", args)
    return LISTP_COST, args.true if args.first().listp() else args.false


def op_raise(args: SExp) -> Never:
    if args.list_len() == 1 and not args.first().listp():
        raise EvalError("klvm raise", args.first())
    else:
        raise EvalError("klvm raise", args)


def op_eq(args: SExp) -> Tuple[int, SExp]:
    if args.list_len() != 2:
        raise EvalError("= takes exactly 2 arguments", args)
    a0 = args.first()
    a1 = args.rest().first()
    if a0.pair or a1.pair:
        raise EvalError("= on list", a0 if a0.pair else a1)
    b0 = a0.as_atom()
    assert b0 is not None
    b1 = a1.as_atom()
    assert b1 is not None
    cost = EQ_BASE_COST
    cost += (len(b0) + len(b1)) * EQ_COST_PER_BYTE
    return cost, (args.true if b0 == b1 else args.false)



from __future__ import annotations

from typing import Any, Callable, List, Tuple, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from klvm.SExp import SExp

OpCallable = Callable[["OpStackType", "ValStackType"], None]

PythonReturnType = Union[
    bytes, Tuple["PythonReturnType", "PythonReturnType"], List["PythonReturnType"]
]

ValType = Union["SExp", PythonReturnType]
ValStackType = List[ValType]

OpStackType = List[OpCallable]


def _roll(op_stack: OpStackType, val_stack: ValStackType) -> None:
    v1 = val_stack.pop()
    v2 = val_stack.pop()
    val_stack.append(v1)
    val_stack.append(v2)


MakeTupleValStackType = List[
    Union[bytes, Tuple[object, object], "MakeTupleValStackType"]
]


def _make_tuple(op_stack: OpStackType, val_stack: ValStackType) -> None:
    left: PythonReturnType = val_stack.pop()  # type: ignore[assignment]
    right: PythonReturnType = val_stack.pop()  # type: ignore[assignment]
    if right == b"":
        val_stack.append([left])
    elif isinstance(right, list):
        v = [left] + right
        val_stack.append(v)
    else:
        val_stack.append((left, right))


def _as_python(op_stack: OpStackType, val_stack: ValStackType) -> None:
    t: SExp = val_stack.pop()  # type: ignore[assignment]
    pair = t.as_pair()
    if pair:
        left, right = pair
        op_stack.append(_make_tuple)
        op_stack.append(_as_python)
        op_stack.append(_roll)
        op_stack.append(_as_python)
        val_stack.append(left)
        val_stack.append(right)
    else:
        # we know that t.atom is not None here because the pair is None
        val_stack.append(t.atom)  # type:ignore[arg-type]


def as_python(sexp: SExp) -> Any:
    op_stack: OpStackType = [_as_python]
    val_stack: ValStackType = [sexp]
    while op_stack:
        op_f = op_stack.pop()
        op_f(op_stack, val_stack)
    return val_stack[-1]
